"""Generate random signals with phase-amplitude coupling inside."""
import logging

import numpy as np

from tensorpac.spectral import morlet

logger = logging.getLogger('tensorpac')


def pac_signals_wavelet(f_pha=10., f_amp=100., sf=1024., n_times=4000.,
                        n_epochs=10, noise=.1, pp=0., rnd_state=0):
    """Generate artificially phase-amplitude coupled signals using wavelets.

    This function is inspired by the code of the pactools toolbox developped by
    Tom Dupre la Tour :cite:`la2017non`.

    Parameters
    ----------
    f_pha : float | 10.
        Frequency for phase. Use either a float number for a centered frequency
        of a band (like [5, 7]) for a bandwidth.
    f_amp : float | 100.
        Frequency for amplitude. Use either a float number for a centered
        frequency of a band (like [60, 80]) for a bandwidth.
    sf : float | 1024.
        Sampling frequency.
    n_times : int | 4000
        Number of time points.
    n_epochs : int | 10
        Number of trials in the dataset.
    noise : float | .1
        Amount of white noise.
    pp : float | 0.
        The preferred-phase of the coupling.
    rnd_state: int | 0
        Fix random of the machine (for reproducibility)

    Returns
    -------
    data : array_like
        Array of signals of shape (n_epochs, n_times).
    time : array_like
        Time vector of shape (n_times,).
    """
    n_times = int(n_times)
    sf = float(sf)
    f_pha, f_amp = np.asarray(f_pha).mean(), np.asarray(f_amp).mean()
    # time = np.mgrid[0:n_epochs, 0:n_times][1] / sf
    time = np.arange(n_times) / sf
    # Random state of the machine :
    rng = np.random.RandomState(rnd_state)
    # Get complex decomposition of random points in the phase frequency band :
    driver = morlet(rng.randn(n_epochs, n_times), sf, f_pha)
    driver /= np.max(driver, axis=1, keepdims=True)
    # Create amplitude signals :
    xh = np.sin(2 * np.pi * f_amp * time.reshape(1, -1))
    dpha = np.exp(-1j * pp)
    modulation = 1. / (1. + np.exp(- 6. * 1. * np.real(driver * dpha)))
    # Modulate the amplitude :
    xh = xh * modulation
    # Get the phase signal :
    xl = np.real(driver)
    # Build the pac signal :
    data = xh + xl + noise * rng.randn(*xh.shape)

    return data, time


def pac_signals_tort(f_pha=10., f_amp=100., sf=1024, n_times=4000, n_epochs=10,
                     chi=0., noise=1., dpha=0., damp=0., rnd_state=0):
    """Generate artificially phase-amplitude coupled signals.

    This function uses the definition of Tort et al. 2010
    :cite:`tort2010measuring`.

    Parameters
    ----------
    f_pha : float | 10.
        Frequency for phase. Use either a float number for a centered frequency
        of a band (like [5, 7]) for a bandwidth.
    f_amp : float | 100.
        Frequency for amplitude. Use either a float number for a centered
        frequency of a band (like [60, 80]) for a bandwidth.
    sf : int | 1024
        Sampling frequency
    n_epochs : int | 10
        Number of datasets
    n_times : int | 4000
        Number of points for each signal.
    chi : float | 0.
        Amount of coupling. If chi=0, signals of phase and amplitude
        are strongly coupled (0.<=chi<=1.).
    noise : float | 1.
        Amount of noise (0<=noise<=3).
    dpha : float | 0.
        Random incertitude on phase frequences (0<=dpha<=100). If f_pha is 2,
        and dpha is 50, the frequency for the phase signal will be between :
        [2-0.5*2, 2+0.5*2]=[1,3]
    damp : float | 0.
        Random incertitude on amplitude frequencies (0<=damp<=100). If f_amp is
        60, and damp is 10, the frequency for the amplitude signal will be
        between : [60-0.1*60, 60+0.1*60]=[54,66]
    rnd_state: int | 0
        Fix random of the machine (for reproducibility)

    Returns
    -------
    data : array_like
        Array of signals of shape (n_epochs, n_channels, n_times).
    time : array_like
        Time vector of shape (n_times,).
    """
    n_times, sf = int(n_times), float(sf)
    # Check the inputs variables :
    chi = 0 if not 0 <= chi <= 1 else chi
    noise = 0 if not 0 <= noise <= 3 else noise
    dpha = 0 if not 0 <= dpha <= 100 else dpha
    damp = 0 if not 0 <= damp <= 100 else damp
    f_pha, f_amp = np.asarray(f_pha), np.asarray(f_amp)
    # time = np.mgrid[0:n_epochs, 0:n_times][1] / sf
    time = np.arange(n_times) / sf
    # Random state of the machine :
    rng = np.random.RandomState(rnd_state)
    # Band / Delta parameters :
    sh = (n_epochs, 1)
    if f_pha.ndim == 0:
        apha = [f_pha * (1. - dpha / 100.), f_pha * (1. + dpha / 100.)]
        del_pha = apha[0] + (apha[1] - apha[0]) * rng.rand(*sh)
    elif f_pha.ndim == 1:
        del_pha = rng.uniform(f_pha[0], f_pha[1], (n_epochs, 1))
    if f_amp.ndim == 0:
        a_amp = [f_amp * (1. - damp / 100.), f_amp * (1. + damp / 100.)]
        del_amp = a_amp[0] + (a_amp[1] - a_amp[0]) * rng.rand(*sh)
    elif f_amp.ndim == 1:
        del_amp = rng.uniform(f_amp[0], f_amp[1], (n_epochs, 1))

    # Create phase and amplitude signals :
    xl = np.sin(2 * np.pi * del_pha * time.reshape(1, -1))
    xh = np.sin(2 * np.pi * del_amp * time.reshape(1, -1))

    # Create the coupling :
    ah = .5 * ((1. - chi) * xl + 1. + chi)
    al = 1.

    # Generate datasets :
    data = (ah * xh) + (al * xl)
    data += noise * rng.rand(*data.shape)  # Add noise

    return data, time



"""Multi-dimentional Gaussian copula mutual information estimation."""
import numpy as np
from scipy.special import psi, ndtri

from tensorpac.config import CONFIG


def ctransform(x):
    """Copula transformation (empirical CDF).

    Parameters
    ----------
    x : array_like
        Array of data. The trial axis should be the last one

    Returns
    -------
    xr : array_like
        Empirical CDF value along the last axis of x. Data is ranked and scaled
        within [0 1] (open interval)
    """
    xi = np.argsort(x)
    xr = np.argsort(xi).astype(float)
    xr += 1.
    xr /= float(xr.shape[-1] + 1)
    return xr


def _copnorm(x):
    return ndtri(ctransform(x))


def copnorm(x):
    """Copula normalization.

    Parameters
    ----------
    x : array_like
        Array of data. The trial axis should be the last one

    Returns
    -------
    cx : array_like
        Standard normal samples with the same empirical CDF value as the input.
        Operates along the last axis
    """
    return np.apply_along_axis(_copnorm, -1, x)


def nd_mi_gg(x, y):
    """Multi-dimentional MI between two Gaussian variables in bits.

    Parameters
    ----------
    x, y : array_like
        Arrays to consider for computing the Mutual Information. The two input
        variables x and y should have the same shape except on the mvaxis
        (if needed).

    Returns
    -------
    mi : array_like
        The mutual information between x and y
    """
    # Multi-dimentional shape checking
    # x.shape (..., x_mvaxis, traxis)
    # y.shape (..., y_mvaxis, traxis)
    ntrl = x.shape[-1]
    nvarx, nvary = x.shape[-2], y.shape[-2]
    nvarxy = nvarx + nvary

    # joint variable along the mvaxis
    xy = np.concatenate((x, y), axis=-2)
    if CONFIG['MI_DEMEAN']:
        xy -= xy.mean(axis=-1, keepdims=True)
    cxy = np.einsum('...ij, ...kj->...ik', xy, xy)
    cxy /= float(ntrl - 1.)

    # submatrices of joint covariance
    cx = cxy[..., :nvarx, :nvarx]
    cy = cxy[..., nvarx:, nvarx:]

    # Cholesky decomposition
    chcxy = np.linalg.cholesky(cxy)
    chcx = np.linalg.cholesky(cx)
    chcy = np.linalg.cholesky(cy)

    # entropies in nats
    # normalizations cancel for mutual information
    hx = np.log(np.einsum('...ii->...i', chcx)).sum(-1)
    hy = np.log(np.einsum('...ii->...i', chcy)).sum(-1)
    hxy = np.log(np.einsum('...ii->...i', chcxy)).sum(-1)

    ln2 = np.log(2)
    if CONFIG['MI_BIASCORRECT']:
        vec = np.arange(1, nvarxy + 1)
        psiterms = psi((ntrl - vec).astype(np.float) / 2.0) / 2.0
        dterm = (ln2 - np.log(ntrl - 1.0)) / 2.0
        hx = hx - nvarx * dterm - psiterms[:nvarx].sum()
        hy = hy - nvary * dterm - psiterms[:nvary].sum()
        hxy = hxy - nvarxy * dterm - psiterms[:nvarxy].sum()

    # MI in bits
    i = (hx + hy - hxy) / ln2
    return i



"""Test gcmi related functions."""
import numpy as np

from tensorpac.gcmi import copnorm, nd_mi_gg
from tensorpac.config import CONFIG


class TestGcmi(object):
    """Test GCMI functions."""

    def test_copnorm(self):
        """Test Gaussian-copula normalization."""
        x = np.random.rand(10, 20, 30)
        copnorm(x)

    def test_mi(self):
        """Test computing MI."""
        rng = np.random.RandomState(0)
        x = rng.rand(100, 1, 100)
        y = rng.rand(100, 1, 100)
        # basic config
        nd_mi_gg(x, y)
        # modified config
        CONFIG['MI_DEMEAN'] = True
        CONFIG['MI_BIASCORRECT'] = True
        nd_mi_gg(x, y)
        CONFIG['MI_BIASCORRECT'] = False
        CONFIG['MI_DEMEAN'] = False



"""Visualization functions."""
import numpy as np
import logging

logger = logging.getLogger('tensorpac')


class _PacVisual(object):
    """Main PAC plotting class."""

    def __init__(self):
        """Init."""
        self._autovmM = True

    def pacplot(self, pac, xvec, yvec, xlabel='', ylabel='', cblabel='',
                title='', fz_labels=12, fz_title=13, fz_cblabel=12,
                cmap='viridis', vmin=None, vmax=None, under=None, over=None,
                bad=None, pvalues=None, p=0.05, interp=None, rmaxis=False,
                dpaxis=False, plotas='imshow', ncontours=5, levels=None,
                levelcmap='Reds', polar=False, colorbar=True, y=1.02,
                subplot=111):
        """Main plotting pac function.

        This method can be used to plot any 2D array.

        Parameters
        ----------
        pac : array_like
            A 2D array.
        xvec : array_like
            The vector to use for the x-axis.
        yvec : array_like
            The vector to use for the y-axis.
        xlabel : string | ''
            Label for the x-axis.
        ylabel : string | ''
            Label for the y-axis.
        cblabel : string | ''
            Label for the colorbar.
        title : string | ''
            Title of the plot.
        fz_labels : float | 12
            Font size of the y- and x-labels
        fz_title : float | 13
            Font size of the title
        fz_cblabel : float | 12
            Font size of the colorbar label
        y : float | 1.02
            Title location.
        cmap : string | 'viridis'
            Name of one Matplotlib's colomap.
        vmin : float | None
            Threshold under which set the color to the uner parameter.
        vmax : float | None
            Threshold over which set the color in the over parameter.
        under : string | 'gray'
            Color for values under the vmin parameter.
        over : string | 'red'
            Color for values over the vmax parameter.
        bad : string | None
            Color for non-significant values.
        pvalues : array_like | None
            P-values to use for masking PAC values. The shape of this
            parameter must be the same as the shape as pac.
        p : float | .05
            If pvalues is pass, use this threshold for masking
            non-significant PAC.
        interp : tuple | None
            Tuple for controlling the 2D interpolation. For example,
            (.1, .1) will multiply the number of row and columns by 10.
        rmaxis : bool | False
            Remove unecessary axis.
        dpaxis : bool | False
            Despine axis.
        plotas : {'imshow', 'contour', 'pcolor'}
            Choose how to display the comodulogram, either using imshow
            ('imshow') or contours ('contour'). If you choose 'contour',
            use the ncontours parameter for controlling the number of
            contours.
        ncontours : int | 5
            Number of contours if plotas is 'contour'.
        levels : list | None
            Add significency levels. This parameter must be a sorted list
            of p-values to use as levels.
        levelcmap : string | Reds
            Colormap of signifiency levels.

        Returns
        -------
        gca: axes
            The current matplotlib axes.
        """
        # Check if pac is 2 dimensions :
        if pac.ndim is not 2:
            raise ValueError("The PAC variable must have two dimensions.")
        # Try import matplotlib :
        try:
            import matplotlib.pyplot as plt
        except:
            raise ValueError("Matplotlib not installed.")
        # Define p-values (if needed) :
        if pvalues is None:
            pvalues = np.zeros_like(pac)
        # 2D interpolation (if needed)
        if interp is not None:
            pac, xvec, yvec = mapinterpolation(pac, xvec, yvec,
                                               interp[0], interp[1])
            pvalues = mapinterpolation(pvalues, self.xvec, self.yvec,
                                       interp[0], interp[1])[0]
        pac = np.ma.masked_array(pac, mask=pvalues >= p)
        # Polar plot :
        if polar:
            plotas = 'pcolor'
            plt.subplot(subplot, projection='polar')
        # Check vmin / vmax
        if (vmin is None) and (vmax is None) and self._autovmM:
            vmin, vmax = min(0, np.nanmin(pac)), max(0, np.nanmax(pac))
            if vmin < 0 and vmax > 0:
                vmax = max(vmax, -vmin)
                vmin = -vmax
        # Plot type :
        toplot = pac.data if levels is not None else pac
        if plotas is 'imshow':
            im = plt.imshow(toplot, aspect='auto', cmap=cmap, origin='upper',
                            vmin=vmin, vmax=vmax, interpolation='none',
                            extent=[xvec[0], xvec[-1], yvec[-1], yvec[0]])
            plt.gca().invert_yaxis()
        elif plotas is 'contour':
            im = plt.contourf(xvec, yvec, toplot, ncontours, cmap=cmap,
                              vmin=vmin, vmax=vmax)
        elif plotas is 'pcolor':
            im = plt.pcolormesh(xvec, yvec, toplot, cmap=cmap, vmin=vmin,
                                vmax=vmax, antialiased=True)
        else:
            raise ValueError("The plotas parameter must either be 'imshow' or "
                             "'contour'")
        self._plt_im = im

        # Add levels :
        if levels is not None:
            levels = [levels] if isinstance(levels, (int, float)) else levels
            plt.contour(pvalues, extent=[xvec[0], xvec[-1], yvec[0], yvec[-1]],
                        levels=levels, cmap=levelcmap)

        # Under/Over/Bad :
        if under is not None:
            im.cmap.set_under(color=under)
        if over is not None:
            im.cmap.set_over(color=over)
        if bad is not None:
            im.cmap.set_bad(color=bad)

        # Title/Xlabel/Ylabel :
        plt.axis('tight')
        plt.xlabel(xlabel, fontsize=fz_labels)
        plt.ylabel(ylabel, fontsize=fz_labels)
        plt.title(title, y=y, fontsize=fz_title)
        plt.clim(vmin=vmin, vmax=vmax)

        # Colorbar
        if colorbar:
            cb = plt.colorbar(im, shrink=0.7, pad=0.01, aspect=10)
            cb.set_label(cblabel, fontsize=fz_cblabel)
            cb.outline.set_visible(False)
        ax = plt.gca()

        # Remove axis :
        if rmaxis:
            for loc, spine in ax.spines.items():
                if loc in ['top', 'right']:
                    spine.set_color('none')
                    ax.tick_params(**{loc: False})

        # Despine axis :
        if dpaxis:
            for loc, spine in ax.spines.items():
                if loc in ['left', 'bottom']:
                    spine.set_position(('outward', 10))
                    spine.set_smart_bounds(True)

        if polar:
            ax.grid(True)
            ax.set_rlabel_position(0)
            ax.set_thetamin(-180)
            ax.set_thetamax(180)
            ax.set_thetagrids([-135, -90, -45, 0, 45, 90, 135, 180])

        return plt.gca()

    def show(self):
        """Display the figure."""
        import matplotlib.pyplot as plt
        plt.show()

    def savefig(self, filename, dpi=600):
        """Save the figure.

        Parameters
        ----------
        filename : string
            The name of the figure to save.
        dpi : int | 600
            DPI of the figure.
        """
        import matplotlib.pyplot as plt
        plt.savefig(filename, dpi=dpi, bbox_inches='tight')


class _PacPlt(_PacVisual):
    """Plotting class for :class:`tensorpac.Pac`."""

    def comodulogram(self, pac, xlabel='Frequency for phase (hz)',
                     ylabel='Frequency for amplitude (hz)',
                     cblabel='PAC values', **kwargs):
        """Plot PAC using comodulogram.

        Parameters
        ----------
        pac : array_like
            PAC array of shape (namp, pha)
        xlabel : string | 'Frequency for phase (hz)'
            Label for the phase axis.
        ylabel : string | 'Frequency for amplitude (hz)'
            Label for the amplitude axis.
        cblabel : string | 'PAC values'
            Colorbar.
        kwargs : dict
            Further arguments are passed to the pacplot() method.

        Returns
        -------
        gca : axes
            The current matplotlib axes.
        """
        if isinstance(pac, np.ndarray) and (pac.ndim == 3):
            logger.warning("3d pac array has been given as an input. Only 2d "
                           "arrays are supported for plotting. Taking the mean"
                           " across the last dimension")
            pac = pac.mean(-1)
        xvec, yvec = self.xvec, self.yvec
        # Disable automatic vmin/vmax :
        self._autovmM = True
        return self.pacplot(pac, xvec, yvec, xlabel, ylabel, cblabel,
                            **kwargs)

    def triplot(self, pac, fvec, tridx, xlabel='Starting frequency (hz)',
                ylabel='Ending frequency (hz)', cblabel='PAC values',
                bad='lightgray', **kwargs):
        """Triangular plot.

        The triplot method is used to find the [starting, ending] frequency
        either for the phase or for the amplitude.

        Parameters
        ----------
        pac : array_like
            Pac array of shape (namp, npha)
        fvec : array_like
            The frequency vector returned by the pac_trivec function.
        tridx : array_like
            The index vector used to build the triangle. This argument is
            also returned by the pac_trivec function.
        xlabel : string | 'Starting frequency (hz)'
            Label for the phase axis.
        ylabel : string | 'Ending frequency (hz)'
            Label for the amplitude axis.
        cblabel : string | 'PAC values'
            Colorbar.
        bad : string | 'lightgray'
            Color for non-significant values.
        kwargs : dict
            Further arguments are passed to the pacplot() method.

        Returns
        -------
        gca : axes
            The current matplotlib axes.
        """
        pac, tridx = np.squeeze(pac), np.squeeze(tridx)
        # ___________________ CHECKING ___________________
        # Check if pac is a raw vector :
        if pac.ndim is not 1:
            raise ValueError("The PAC variable must be a row vector.")
        if len(pac) != tridx.shape[0]:
            raise ValueError("PAC and tridx variables must have the same "
                             "length.")

        # ___________________ RECONSTRUCT ___________________
        npac = tridx.max() + 2
        rpac = np.zeros((npac, npac), dtype=float)
        for num, k in enumerate(tridx):
            rpac[k[0], k[1]] = pac[num]
        # Build mask :
        mask = np.zeros_like(rpac, dtype=bool)
        mask[np.triu_indices_from(mask)] = True
        # Mask the lower triangle :
        rpac = np.ma.masked_array(np.flipud(rpac), mask=mask)

        # ___________________ PLOT ___________________
        # Define frequency vector :
        vector = fvec[tridx[:, 0] == 0, 0]
        xvec = yvec = np.append(vector, [fvec.max()])
        # Enable automatic vmin/vmax :
        self._autovmM = False
        return self.pacplot(rpac, xvec, yvec, xlabel, ylabel, cblabel,
                            bad=bad, **kwargs)


class _PolarPlt(_PacVisual):
    """Plotting class for :class:`tensorpac.PreferredPhase`."""

    def polar(self, amp, xvec, yvec, interp=None, **kwargs):
        """Polar representation.

        This method is used to visualize amplitude as a function of phase using
        a polar (circle) representation.

        Parameters
        ----------
        amp : array_like
            2D array.
        xvec : array_like
            Vector for the x-axis.
        yvec : array_like
            Vector for the y-axis (phases).
        interp : float | None
            Interplation factor.
        kwargs : dict
            Further arguments are passed to the pacplot() method.

        Returns
        -------
        gca : axes
            The current matplotlib axes.
        """
        # Interpolation :
        if interp is not None:
            amp, yvec, xvec = mapinterpolation(amp, yvec, xvec, interp, 1)
        # Disable automatic vmin/vmax :
        self._autovmM = False
        return self.pacplot(amp, xvec, yvec, polar=True, **kwargs)


def mapinterpolation(data, x=None, y=None, interpx=1, interpy=1):
    """Interpolate a 2D map."""
    # Get data size :
    dim2, dim1 = data.shape
    # Define xticklabel and yticklabel :
    if x is None:
        x = np.arange(0, dim1, interpx)
    if y is None:
        y = np.arange(0, dim2, interpy)
    # Define the meshgrid :
    xi, yi = np.meshgrid(
        np.arange(0, dim1 - 1, interpx), np.arange(0, dim2 - 1, interpy))
    # 2D interpolation :
    datainterp = interp2(data, xi, yi)
    # Linearly interpolate vectors :
    xveci = np.linspace(x[0], x[-1], datainterp.shape[0])
    yveci = np.linspace(y[0], y[-1], datainterp.shape[1])
    return datainterp, xveci, yveci


def interp2(z, xi, yi, extrapval=0):
    """Linear interpolation.

    This function is equivalent to interp2(z, xi, yi,'linear') in Matlab.

    Parameters
    ----------
    z : array_like
        Array to interpolate.
    xi : array_like
        Array of x coordinates where interpolation is required.
    yi : array_like
        Array of y coordinates where interpolation is required.
    extrapval : float | 0.
        Value for out of range positions.

    Returns
    -------
        f: array_like
            Extrapolated data.
    """
    x = xi.copy()
    y = yi.copy()
    nrows, ncols = z.shape

    if nrows < 2 or ncols < 2:
        raise Exception("z shape is too small")

    if not x.shape == y.shape:
        raise Exception("sizes of X indexes and Y-indexes must match")

    # find x values out of range
    x_bad = ((x < 0) | (x > ncols - 1))
    if x_bad.any():
        x[x_bad] = 0

    # find y values out of range
    y_bad = ((y < 0) | (y > nrows - 1))
    if y_bad.any():
        y[y_bad] = 0

    # linear indexing. z must be in 'C' order
    ndx = np.floor(y) * ncols + np.floor(x)
    ndx = ndx.astype('int32')

    # fix parameters on x border
    d = (x == ncols - 1)
    x = (x - np.floor(x))
    if d.any():
        x[d] += 1
        ndx[d] -= 1

    # fix parameters on y border
    d = (y == nrows - 1)
    y = (y - np.floor(y))
    if d.any():
        y[d] += 1
        ndx[d] -= ncols

    # interpolate
    one_minus_t = 1 - y
    z = z.ravel()
    f = (z[ndx] * one_minus_t + z[ndx + ncols] * y) * (1 - x) + (
        z[ndx + 1] * one_minus_t + z[ndx + ncols + 1] * y) * x

    # Set out of range positions to extrapval
    if x_bad.any():
        f[x_bad] = extrapval
    if y_bad.any():
        f[y_bad] = extrapval

    return f



"""Switch and utility functions for PAC methods."""
import numpy as np
from functools import partial


def get_pac_fcn(idp, n_bins, p, implementation="tensor", full=False):
    """Get the function for computing Phase-Amplitude coupling.

    This function also allow to switch between Tensor / Numba implementations
    of most of the functions.
    """
    n_bins, p = np.int64(n_bins), np.float64(p)
    assert implementation in ['tensor', 'numba']
    if implementation is 'tensor':
        from tensorpac.methods.meth_pac import (
            mean_vector_length, modulation_index, heights_ratio,
            norm_direct_pac, phase_locking_value, gauss_cop_pac)
        METH = {
            1: partial(mean_vector_length),
            2: partial(modulation_index, n_bins=n_bins),
            3: partial(heights_ratio, n_bins=n_bins),
            4: partial(norm_direct_pac, p=p),
            5: partial(phase_locking_value),
            6: partial(gauss_cop_pac)}
    elif implementation is 'numba':
        from tensorpac.methods.meth_pac_nb import (
            mean_vector_length_nb, modulation_index_nb, heights_ratio_nb,
            norm_direct_pac_nb, phase_locking_value_nb)
        # Gaussian-copula PAC can't be compiled with Numba. Hence, we force
        # using the Tensor implementation
        from tensorpac.methods.meth_pac import gauss_cop_pac
        METH = {
            1: partial(mean_vector_length_nb),
            2: partial(modulation_index_nb, n_bins=n_bins),
            3: partial(heights_ratio_nb, n_bins=n_bins),
            4: partial(norm_direct_pac_nb, p=p),
            5: partial(phase_locking_value_nb),
            6: partial(gauss_cop_pac)}

    if full:
        return METH
    else:
        return METH[idp]


def pacstr(idpac):
    """Return correspond methods string."""
    # Pac methods :
    if idpac[0] == 1:
        method = 'Mean Vector Length (MVL, Canolty et al. 2006)'
    elif idpac[0] == 2:
        method = 'Modulation Index (MI, Tort et al. 2010)'
    elif idpac[0] == 3:
        method = 'Heights ratio (HR, Lakatos et al. 2005)'
    elif idpac[0] == 4:
        method = 'Normalized Direct Pac (ndPac, Ozkurt et al. 2012)'
    elif idpac[0] == 5:
        method = 'Phase-Locking Value (PLV, Penny et al. 2008)'
    elif idpac[0] == 6:
        method = 'Gaussian Copula PAC (gcPac)'
    else:
        raise ValueError("No corresponding pac method.")

    # Surrogate method :
    if idpac[1] == 0:
        suro = 'No surrogates'
    elif idpac[1] == 1:
        suro = 'Permute phase across trials (Tort et al. 2010)'
    elif idpac[1] == 2:
        suro = 'Swap amplitude time blocks (Bahramisharif et al. 2013)'
    elif idpac[1] == 3:
        suro = 'Time lag (Canolty et al. 2006)'
    else:
        raise ValueError("No corresponding surrogate method.")

    # Normalization methods :
    if idpac[2] == 0:
        norm = 'No normalization'
    elif idpac[2] == 1:
        norm = 'Substract the mean of surrogates'
    elif idpac[2] == 2:
        norm = 'Divide by the mean of surrogates'
    elif idpac[2] == 3:
        norm = 'Substract then divide by the mean of surrogates'
    elif idpac[2] == 4:
        norm = "Substract the mean and divide by the deviation of the " + \
               "surrogates"
    else:
        raise ValueError("No corresponding normalization method.")

    return method, suro, norm



"""
Tensorpac
=========

Tensorpac is an open-source Python toolbox designed for computing
Phase-Amplitude Coupling.
"""
import logging

from tensorpac import methods, signals, utils, stats  # noqa
from tensorpac.pac import (Pac, EventRelatedPac, PreferredPhase)  # noqa
from tensorpac.io import set_log_level
# Set 'info' as the default logging level
logger = logging.getLogger('brainets')
set_log_level('info')

__version__ = "0.6.4"



"""Numba implementation of some PAC functions."""
import numpy as np
from scipy.special import erfinv


# if Numba not installed, this section should return a Numba-free jit wrapper
try:
    import numba
    def jit(signature=None, nopython=True, nogil=True, fastmath=True,  # noqa
            cache=True, **kwargs):
        return numba.jit(signature_or_function=signature, cache=cache,
                         nogil=nogil, fastmath=fastmath, nopython=nopython,
                         **kwargs)
except:
    def jit(*args, **kwargs):  # noqa
        def _jit(func):
            return func
        return _jit


@jit("f8[:,:,:](f8[:,:,:], f8[:,:,:])")
def mean_vector_length_nb(pha, amp):
    """Numba-based Mean Vector Length (MVL).

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, n_epochs, n_times)
        and the array of amplitudes of shape (n_amp, n_epochs, n_times). Both
        arrays should be of type float64 (np.float64)

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, n_epochs)

    References
    ----------
    Canolty et al. 2006 :cite:`canolty2006high`
    """
    n_pha, n_epochs, n_times = pha.shape
    n_amp, _, _ = amp.shape
    pac = np.zeros((n_amp, n_pha, n_epochs), dtype=np.float64)
    # single conversion
    exp_pha = np.exp(1j * pha)
    amp_comp = amp.astype(np.complex128)
    for a in range(n_amp):
        for p in range(n_pha):
            for tr in range(n_epochs):
                _pha = np.ascontiguousarray(exp_pha[p, tr, :])
                _amp = np.ascontiguousarray(amp_comp[a, tr, :])
                pac[a, p, tr] = abs(np.dot(_amp, _pha))
    pac /= n_times

    return pac


@jit("f8[:](f8[:], f8[:], u8, b1)")
def _kl_hr_nb(pha, amp, n_bins=18, mean_bins=True):
    """Binarize the amplitude according to phase values.

    This function is shared by the Kullback-Leibler Distance and the
    Height Ratio.
    """
    vecbin = np.linspace(-np.pi, np.pi, n_bins + 1)
    phad = np.digitize(pha, vecbin) - 1
    u_phad = np.unique(phad)

    abin = np.zeros((len(u_phad)), dtype=np.float64)
    for n_i, i in enumerate(u_phad):
        # find where phase take vecbin values
        idx = np.ascontiguousarray((phad == i).astype(np.float64))
        m = idx.sum() if mean_bins else 1.
        # take the sum of amplitude inside the bin
        abin[n_i] = np.dot(np.ascontiguousarray(amp), idx) / m

    return abin


@jit("f8[:,:,:](f8[:,:,:], f8[:,:,:], u8)")
def modulation_index_nb(pha, amp, n_bins=18):
    """Numba-based Modulation index (MI).

    The modulation index is obtained using the Kullback Leibler Distance which
    measures how much the distribution of binned amplitude differs from a
    uniform distribution.

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, n_epochs, n_times)
        and the array of amplitudes of shape (n_amp, n_epochs, n_times). Both
        arrays should be of type float64 (np.float64)
    n_bins : int | 18
        Number of bins to binarize the amplitude according to phase intervals
        (should be np.int64)

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Tort et al. 2010 :cite:`tort2010measuring`
    """
    n_pha, n_epochs, n_times = pha.shape
    n_amp, _, _ = amp.shape
    pac = np.zeros((n_amp, n_pha, n_epochs), dtype=np.float64)
    bin_log = np.log(n_bins)

    for a in range(n_amp):
        for p in range(n_pha):
            for tr in range(n_epochs):
                # select phase and amplitude
                _pha = np.ascontiguousarray(pha[p, tr, :])
                _amp = np.ascontiguousarray(amp[a, tr, :])
                # get the probability of each amp bin
                p_j = _kl_hr_nb(_pha, _amp, n_bins=n_bins, mean_bins=True)
                p_j /= p_j.sum()
                # log it (only if strictly positive)
                if np.all(p_j > 0.):
                    p_j *= np.log(p_j)
                    # compute the PAC
                    pac[a, p, tr] = 1. + p_j.sum() / bin_log
                else:
                    pac[a, p, tr] = 0.

    return pac


@jit("f8[:,:,:](f8[:,:,:], f8[:,:,:], u8)")
def heights_ratio_nb(pha, amp, n_bins=18):
    """Numba-based Heights ratio (HR).

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, n_epochs, n_times)
        and the array of amplitudes of shape (n_amp, n_epochs, n_times). Both
        arrays should be of type float64 (np.float64)
    n_bins : int | 18
        Number of bins to binarize the amplitude according to phase intervals
        (should be np.int64)

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Lakatos et al. 2005 :cite:`lakatos2005oscillatory`
    """
    n_pha, n_epochs, n_times = pha.shape
    n_amp, _, _ = amp.shape
    pac = np.zeros((n_amp, n_pha, n_epochs), dtype=np.float64)

    for a in range(n_amp):
        for p in range(n_pha):
            for tr in range(n_epochs):
                # select phase and amplitude
                _pha = np.ascontiguousarray(pha[p, tr, :])
                _amp = np.ascontiguousarray(amp[a, tr, :])
                # get the probability of each amp bin
                p_j = _kl_hr_nb(_pha, _amp, n_bins=n_bins, mean_bins=True)
                p_j /= p_j.sum()
                # find (maximum, minimum) of the binned distribution
                h_max, h_min = np.max(p_j), np.min(p_j)
                # compute the PAC
                pac[a, p, tr] = (h_max - h_min) / h_max

    return pac


def phase_locking_value_nb(pha, pha_amp):
    """Numba-based Phase Locking-Value (PLV).

    In order to measure the phase locking value, the phase of the amplitude of
    the higher-frequency signal must be provided, and not the amplitude as in
    most other PAC functions.

    Parameters
    ----------
    pha, pha_amp : array_like
        Respectively the arrays of phases of shape (n_pha, n_epochs, n_times)
        for the lower frequency and the array of phase of the amplitude signal
        of shape (n_pha_amp, n_epochs, n_times) for the higher frequency. Both
        arrays should be of type float64 (np.float64)

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_pha_amp, n_pha, ...)

    References
    ----------
    Penny et al. 2008 :cite:`penny2008testing`, Lachaux et al. 1999
    :cite:`lachaux1999measuring`
    """
    n_pha, n_epochs, n_times = pha.shape
    n_amp, _, _ = pha_amp.shape
    pac = np.zeros((n_amp, n_pha, n_epochs), dtype=np.float64)
    # single conversion
    exp_pha = np.exp(1j * pha)
    exp_pha_amp = np.exp(-1j * pha_amp)
    for a in range(n_amp):
        for p in range(n_pha):
            for tr in range(n_epochs):
                _pha = exp_pha[p, tr, :]
                _pha_amp = exp_pha_amp[a, tr, :]
                pac[a, p, tr] = abs(np.dot(_pha, _pha_amp))
    pac /= n_times

    return pac

"""
I don't think this function can be entirely compiled with Numba because of two
issues :

    * Numba supports the mean / std but not across a specific axis
    * erfinv is a special function of scipy that don't seems to be supported
      for the moment

Therefore, the beginning and the end of the function are tensor-based while the
core function that computes PAC is the Numba compliant MVL.
"""
def norm_direct_pac_nb(pha, amp, p=.05):
    """Numba-based Normalized direct Pac (ndPAC).

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, n_epochs, n_times)
        and the array of amplitudes of shape (n_amp, n_epochs, n_times). Both
        arrays should be of type float64 (np.float64)
    p : float | .05
        P-value to use for thresholding. Sub-threshold PAC values
        will be set to 0. To disable this behavior (no masking), use ``p=1`` or
        ``p=None``. Should be a np.float64

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Ozkurt et al. :cite:`ozkurt2012statistically`
    """
    n_times = pha.shape[-1]
    # z-score normalization to approximate assumptions
    amp = np.subtract(amp, np.mean(amp, axis=-1, keepdims=True))
    amp = np.divide(amp, np.std(amp, ddof=1, axis=-1, keepdims=True))
    # compute pac using MVL (need to remultiply by n_times)
    pac = mean_vector_length_nb(pha, amp) * n_times

    # no thresholding
    if p == 1. or p is None:
        return pac / n_times

    s = pac ** 2
    pac /= n_times
    # set to zero non-significant values
    xlim = n_times * erfinv(1 - p) ** 2
    pac[s <= 2 * xlim] = 0.
    return pac



"""Test I/O."""
import logging

from tensorpac.io import (set_log_level, is_pandas_installed,
                          is_statsmodels_installed, progress_bar)

logger = logging.getLogger('tensorpac')


levels = ['profiler', 'debug', 'info', 'warning', 'error', 'critical']


class TestIO(object):
    """Test statistical functions."""

    def test_log_level(self):
        """Test setting the log level."""
        for l in levels:
            set_log_level(l)
        set_log_level(False)
        set_log_level(True)
        set_log_level(match="ok")
        logger.info("show me ok")
        logger.info("show me")

    def test_logger(self):
        """Test logger levels."""
        set_log_level("profiler")
        logger.profiler("profiler")
        logger.debug("debug")
        logger.info("info")
        logger.warning("warning")
        logger.critical("critical")

    def test_progress_bar(self):
        """Test progress bar."""
        progress_bar(5, 10)

    def test_dependance(self):
        """Test dependancies."""
        is_statsmodels_installed()
        is_pandas_installed()



"""Utility functions."""
import logging

import numpy as np
from scipy.signal import periodogram

from tensorpac.methods.meth_pac import _kl_hr
from tensorpac.pac import _PacObj, _PacVisual
from tensorpac.io import set_log_level

from matplotlib.gridspec import GridSpec
import matplotlib.pyplot as plt

logger = logging.getLogger('tensorpac')


def pac_vec(f_pha='mres', f_amp='mres'):
    """Generate cross-frequency coupling vectors.

    Parameters
    ----------
    Frequency vector for the phase and amplitude. Here you can use
    several forms to define those vectors :

        * Basic list/tuple (ex: [2, 4] or [8, 12]...)
        * List of frequency bands (ex: [[2, 4], [5, 7]]...)
        * Dynamic definition : (start, stop, width, step)
        * Range definition (ex : np.arange(3) => [[0, 1], [1, 2]])
        * Using a string. `f_pha` and `f_amp` can be 'lres', 'mres', 'hres'
          respectively for low, middle and high resolution vectors. In that
          case, it uses the definition proposed by Bahramisharif et al. 2013
          :cite:`bahramisharif2013propagating` i.e
          f_pha = [f - f / 4, f + f / 4] and f_amp = [f - f / 8, f + f / 8]

    Returns
    -------
    f_pha, f_amp : array_like
        Arrays containing the pairs of phase and amplitude frequencies. Each
        vector have a shape of (N, 2).
    """
    nb_fcy = dict(lres=10, mres=30, hres=50, demon=70, hulk=100)
    if isinstance(f_pha, str):
        # get where phase frequencies start / finish / number
        f_pha_start, f_pha_end = 2, 20
        f_pha_nb = nb_fcy[f_pha]
        # f_pha = [f - f / 4, f + f / 4]
        f_pha_mid = np.linspace(f_pha_start, f_pha_end, f_pha_nb)
        f_pha = np.c_[f_pha_mid - f_pha_mid / 4., f_pha_mid + f_pha_mid / 4.]
    if isinstance(f_amp, str):
        # get where amplitude frequencies start / finish / number
        f_amp_start, f_amp_end = 60, 160
        f_amp_nb = nb_fcy[f_amp]
        # f_amp = [f - f / 8, f + f / 8]
        f_amp_mid = np.linspace(f_amp_start, f_amp_end, f_amp_nb)
        f_amp = np.c_[f_amp_mid - f_amp_mid / 8., f_amp_mid + f_amp_mid / 8.]

    return _check_freq(f_pha), _check_freq(f_amp)


def _check_freq(f):
    """Check the frequency definition."""
    f = np.atleast_2d(np.asarray(f))
    #
    if len(f.reshape(-1)) == 1:
        raise ValueError("The length of f should at least be 2.")
    elif 2 in f.shape:  # f of shape (N, 2) or (2, N)
        if f.shape[1] is not 2:
            f = f.T
    elif np.squeeze(f).shape == (4,):  # (f_start, f_end, f_width, f_step)
        f = _pair_vectors(*tuple(np.squeeze(f)))
    else:  # Sequential
        f = f.reshape(-1)
        f.sort()
        f = np.c_[f[0:-1], f[1::]]
    return f


def _pair_vectors(f_start, f_end, f_width, f_step):
    # Generate two array for phase and amplitude :
    fdown = np.arange(f_start, f_end - f_width, f_step)
    fup = np.arange(f_start + f_width, f_end, f_step)
    return np.c_[fdown, fup]


def pac_trivec(f_start=60., f_end=160., f_width=10.):
    """Generate triangular vector.

    By contrast with the pac_vec function, this function generate frequency
    vector with an increasing frequency bandwidth.

    Parameters
    ----------
    f_start : float | 60.
        Starting frequency.
    f_end : float | 160.
        Ending frequency.
    f_width : float | 10.
        Frequency bandwidth increase between each band.

    Returns
    -------
    f : array_like
        The triangular vector.
    tridx : array_like
        The triangular index for the reconstruction.
    """
    starting = np.arange(f_start, f_end + f_width, f_width)
    f, tridx = np.array([]), np.array([])
    for num, k in enumerate(starting[0:-1]):
        # Lentgh of the vector to build :
        le = len(starting) - (num + 1)
        # Create the frequency vector for this starting frequency :
        fst = np.c_[np.full(le, k), starting[num + 1::]]
        nfst = fst.shape[0]
        # Create the triangular index for this vector of frequencies :
        idx = np.c_[np.flipud(np.arange(nfst)), np.full(nfst, num)]
        tridx = np.concatenate((tridx, idx), axis=0) if tridx.size else idx
        f = np.concatenate((f, fst), axis=0) if f.size else fst
    return f, tridx


class PSD(object):
    """Power Spectrum Density for electrophysiological brain data.

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_epochs, n_times)
    sf : float
        The sampling frequency.
    """

    def __init__(self, x, sf):
        """Init."""
        assert isinstance(x, np.ndarray) and (x.ndim == 2), (
            "x should be a 2d array of shape (n_epochs, n_times)")
        self._n_trials, self._n_times = x.shape
        logger.info(f"Compute PSD over {self._n_trials} trials and "
                    f"{self._n_times} time points")
        self._freqs, self._psd = periodogram(x, fs=sf, window=None,
                                             nfft=self._n_times,
                                             detrend='constant',
                                             return_onesided=True,
                                             scaling='density', axis=1)

    def plot(self, f_min=None, f_max=None, confidence=95, interp=None,
             log=False, grid=True, fz_title=18, fz_labels=15):
        """Plot the PSD.

        Parameters
        ----------
        f_min, f_max : (int, float) | None
            Frequency bounds to use for plotting
        confidence : (int, float) | None
            Light gray confidence interval. If None, no interval will be
            displayed
        interp : int | None
            Line interpolation integer. For example, if interp is 10 the number
            of points is going to be multiply by 10
        log : bool | False
            Use a log scale representation
        grid : bool | True
            Add a grid to the plot
        fz_title : int | 18
            Font size for the title
        fz_labels : int | 15
            Font size the x/y labels

        Returns
        -------
        ax : Matplotlib axis
            The matplotlib axis that contains the figure
        """
        import matplotlib.pyplot as plt
        f_types = (int, float)
        # interpolation
        xvec, yvec = self._freqs, self._psd
        if isinstance(interp, int) and (interp > 1):
            # from scipy.interpolate import make_interp_spline, BSpline
            from scipy.interpolate import interp1d
            xnew = np.linspace(xvec[0], xvec[-1], len(xvec) * interp)
            f = interp1d(xvec, yvec, kind='quadratic', axis=1)
            yvec = f(xnew)
            xvec = xnew
        # (f_min, f_max)
        f_min = xvec[0] if not isinstance(f_min, f_types) else f_min
        f_max = xvec[-1] if not isinstance(f_max, f_types) else f_max
        # plot main psd
        plt.plot(xvec, yvec.mean(0), color='black',
                 label='mean PSD over trials')
        # plot confidence interval
        if isinstance(confidence, (int, float)) and (0 < confidence < 100):
            logger.info(f"    Add {confidence}th confidence interval")
            interval = (100. - confidence) / 2
            kw = dict(axis=0, interpolation='nearest')
            psd_min = np.percentile(yvec, interval, **kw)
            psd_max = np.percentile(yvec, 100. - interval, **kw)
            plt.fill_between(xvec, psd_max, psd_min, color='lightgray',
                             alpha=0.5,
                             label=f"{confidence}th confidence interval")
            plt.legend(fontsize=fz_labels)
        plt.xlabel("Frequencies (Hz)", fontsize=fz_labels)
        plt.ylabel("Power (V**2/Hz)", fontsize=fz_labels)
        plt.title(f"PSD mean over {self._n_trials} trials", fontsize=fz_title)
        plt.xlim(f_min, f_max)
        if log:
            from matplotlib.ticker import ScalarFormatter
            plt.xscale('log', basex=10)
            plt.gca().xaxis.set_major_formatter(ScalarFormatter())
        if grid:
            plt.grid(color='grey', which='major', linestyle='-',
                     linewidth=1., alpha=0.5)
            plt.grid(color='lightgrey', which='minor', linestyle='--',
                     linewidth=0.5, alpha=0.5)

        return plt.gca()

    def plot_st_psd(self, f_min=None, f_max=None, log=False, grid=True,
                    fz_title=18, fz_labels=15, fz_cblabel=15, **kw):
        """Single-trial PSD plot.

        Parameters
        ----------
        f_min, f_max : (int, float) | None
            Frequency bounds to use for plotting
        log : bool | False
            Use a log scale representation
        grid : bool | True
            Add a grid to the plot
        fz_title : int | 18
            Font size for the title
        fz_labels : int | 15
            Font size the x/y labels
        fz_cblabel : int | 15
            Font size the colorbar label labels

        Returns
        -------
        ax : Matplotlib axis
            The matplotlib axis that contains the figure
        """
        # manage input variables
        kw['fz_labels'] = kw.get('fz_labels', fz_labels)
        kw['fz_title'] = kw.get('fz_title', fz_title)
        kw['fz_cblabel'] = kw.get('fz_cblabel', fz_title)
        kw['xlabel'] = kw.get('xlabel', "Frequencies (Hz)")
        kw['ylabel'] = kw.get('ylabel', "Trials")
        kw['title'] = kw.get('title', "Single-trial PSD")
        kw['cblabel'] = kw.get('cblabel', "Power (V**2/Hz)")
        # (f_min, f_max)
        xvec, psd = self._freqs, self._psd
        f_types = (int, float)
        f_min = xvec[0] if not isinstance(f_min, f_types) else f_min
        f_max = xvec[-1] if not isinstance(f_max, f_types) else f_max
        # locate (f_min, f_max) indices
        f_min_idx = np.abs(xvec - f_min).argmin()
        f_max_idx = np.abs(xvec - f_max).argmin()
        sl_freq = slice(f_min_idx, f_max_idx)
        xvec = xvec[sl_freq]
        psd = psd[:, sl_freq]
        # make the 2D plot
        _viz = _PacVisual()
        trials = np.arange(self._n_trials)
        _viz.pacplot(psd, xvec, trials, **kw)
        if log:
            from matplotlib.ticker import ScalarFormatter
            plt.xscale('log', basex=10)
            plt.gca().xaxis.set_major_formatter(ScalarFormatter())
        if grid:
            plt.grid(color='grey', which='major', linestyle='-',
                     linewidth=1., alpha=0.5)
            plt.grid(color='lightgrey', which='minor', linestyle='--',
                     linewidth=0.5, alpha=0.5)

        return plt.gca()

    def show(self):
        """Display the PSD figure."""
        import matplotlib.pyplot as plt
        plt.show()

    @property
    def freqs(self):
        """Get the frequency vector."""
        return self._freqs

    @property
    def psd(self):
        """Get the psd value."""
        return self._psd


class BinAmplitude(_PacObj):
    """Bin the amplitude according to the phase.

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_epochs, n_times)
    sf : float
        The sampling frequency
    f_pha : tuple, list | [2, 4]
        List of two floats describing the frequency bounds for extracting the
        phase
    f_amp : tuple, list | [60, 80]
        List of two floats describing the frequency bounds for extracting the
        amplitude
    n_bins : int | 18
        Number of bins to use to binarize the phase and the amplitude
    dcomplex : {'wavelet', 'hilbert'}
        Method for the complex definition. Use either 'hilbert' or
        'wavelet'.
    cycle : tuple | (3, 6)
        Control the number of cycles for filtering (only if dcomplex is
        'hilbert'). Should be a tuple of integers where the first one
        refers to the number of cycles for the phase and the second for the
        amplitude :cite:`bahramisharif2013propagating`.
    width : int | 7
        Width of the Morlet's wavelet.
    edges : int | None
        Number of samples to discard to avoid edge effects due to filtering
    """

    def __init__(self, x, sf, f_pha=[2, 4], f_amp=[60, 80], n_bins=18,
                 dcomplex='hilbert', cycle=(3, 6), width=7, edges=None,
                 n_jobs=-1):
        """Init."""
        _PacObj.__init__(self, f_pha=f_pha, f_amp=f_amp, dcomplex=dcomplex,
                         cycle=cycle, width=width)
        # check
        x = np.atleast_2d(x)
        assert x.ndim <= 2, ("`x` input should be an array of shape "
                             "(n_epochs, n_times)")
        assert isinstance(sf, (int, float)), ("`sf` input should be a integer "
                                              "or a float")
        assert all([isinstance(k, (int, float)) for k in f_pha]), (
            "`f_pha` input should be a list of two integers / floats")
        assert all([isinstance(k, (int, float)) for k in f_amp]), (
            "`f_amp` input should be a list of two integers / floats")
        assert isinstance(n_bins, int), "`n_bins` should be an integer"
        logger.info(f"Binning {f_amp}Hz amplitude according to {f_pha}Hz "
                    "phase")
        # extract phase and amplitude
        kw = dict(keepfilt=False, edges=edges, n_jobs=n_jobs)
        pha = self.filter(sf, x, 'phase', **kw)
        amp = self.filter(sf, x, 'amplitude', **kw)
        # binarize amplitude according to phase
        self._amplitude = _kl_hr(pha, amp, n_bins, mean_bins=False).squeeze()
        self.n_bins = n_bins

    def plot(self, unit='rad', normalize=False, **kw):
        """Plot the amplitude.

        Parameters
        ----------
        unit : {'rad', 'deg'}
            The unit to use for the phase. Use either 'deg' for degree or 'rad'
            for radians
        normalize : bool | None
            Normalize the histogram by the maximum
        kw : dict | {}
            Additional inputs are passed to the matplotlib.pyplot.bar function

        Returns
        -------
        ax : Matplotlib axis
            The matplotlib axis that contains the figure
        """
        import matplotlib.pyplot as plt
        assert unit in ['rad', 'deg']
        if unit == 'rad':
            self._phase = np.linspace(-np.pi, np.pi, self.n_bins)
            width = 2 * np.pi / self.n_bins
        elif unit == 'deg':
            self._phase = np.linspace(-180, 180, self.n_bins)
            width = 360 / self.n_bins
        amp_mean = self._amplitude.mean(1)
        if normalize:
            amp_mean /= amp_mean.max()
        plt.bar(self._phase, amp_mean, width=width, **kw)
        plt.xlabel(f"Frequency phase ({self.n_bins} bins)", fontsize=18)
        plt.ylabel("Amplitude", fontsize=18)
        plt.title("Binned amplitude")
        plt.autoscale(enable=True, axis='x', tight=True)

    def show(self):
        """Show the figure."""
        import matplotlib.pyplot as plt
        plt.show()

    @property
    def amplitude(self):
        """Get the amplitude value."""
        return self._amplitude

    @property
    def phase(self):
        """Get the phase value."""
        return self._phase


class ITC(_PacObj, _PacVisual):
    """Compute the Inter-Trials Coherence (ITC).

    The Inter-Trials Coherence (ITC) is a measure of phase consistency over
    trials for a single recording site (electrode / sensor etc.).

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_epochs, n_times)
    sf : float
        The sampling frequency
    f_pha : tuple, list | [2, 4]
        List of two floats describing the frequency bounds for extracting the
        phase
    dcomplex : {'wavelet', 'hilbert'}
        Method for the complex definition. Use either 'hilbert' or
        'wavelet'.
    cycle : tuple | 3
        Control the number of cycles for filtering the phase (only if dcomplex
        is 'hilbert').
    width : int | 7
        Width of the Morlet's wavelet.
    edges : int | None
        Number of samples to discard to avoid edge effects due to filtering
    """

    def __init__(self, x, sf, f_pha=[2, 4], dcomplex='hilbert', cycle=3,
                 width=7, edges=None, n_jobs=-1, verbose=None):
        """Init."""
        set_log_level(verbose)
        _PacObj.__init__(self, f_pha=f_pha, f_amp=[60, 80], dcomplex=dcomplex,
                         cycle=(cycle, 6), width=width)
        _PacVisual.__init__(self)
        # check
        x = np.atleast_2d(x)
        assert x.ndim <= 2, ("`x` input should be an array of shape "
                             "(n_epochs, n_times)")
        self._n_trials = x.shape[0]
        logger.info("Inter-Trials Coherence (ITC)")
        logger.info(f"    extracting {len(self.xvec)} phases")
        # extract phase and amplitude
        kw = dict(keepfilt=False, edges=edges, n_jobs=n_jobs)
        pha = self.filter(sf, x, 'phase', **kw)
        # compute itc
        self._itc = np.abs(np.exp(1j * pha).mean(1)).squeeze()
        self._sf = sf

    def plot(self, times=None, **kw):
        """Plot the Inter-Trials Coherence.

        Parameters
        ----------
        times : array_like | None
            Custom time vector to use
        kw : dict | {}
            Additional inputs are either pass to the matplotlib.pyplot.plot
            function if a single phase band is used, otherwise to the
            matplotlib.pyplot.pcolormesh function

        Returns
        -------
        ax : Matplotlib axis
            The matplotlib axis that contains the figure
        """
        import matplotlib.pyplot as plt
        n_pts = self._itc.shape[-1]
        if not isinstance(times, np.ndarray):
            times = np.arange(n_pts) / self._sf
        times = times[self._edges]
        assert len(times) == n_pts, ("The length of the time vector should be "
                                     "{n_pts}")
        xlab = 'Time'
        title = f"Inter-Trials Coherence ({self._n_trials} trials)"
        if self._itc.ndim == 1:
            plt.plot(times, self._itc, **kw)
        elif self._itc.ndim == 2:
            vmin = kw.get('vmin', np.percentile(self._itc, 1))
            vmax = kw.get('vmax', np.percentile(self._itc, 99))
            self.pacplot(self._itc, times, self.xvec, vmin=vmin, vmax=vmax,
                         ylabel="Frequency for phase (Hz)", xlabel=xlab,
                         title=title, **kw)
        return plt.gca()

    def show(self):
        """Show the figure."""
        import matplotlib.pyplot as plt
        plt.show()

    @property
    def itc(self):
        """Get the itc value."""
        return self._itc


class PeakLockedTF(_PacObj, _PacVisual):
    """Peak-Locked Time-frequency representation.

    This class can be used in order to re-align time-frequency representations
    around a time-point (cue) according to the closest phase peak. This type
    of visualization can bring out a cyclic behavior of the amplitude at a
    given phase, potentially indicating the presence of a phase-amplitude
    coupling. Here's the detailed pipeline :

        * Filter around a single phase frequency bands and across multiple
          amplitude frequencies
        * Use a `cue` which define the time-point to use for the realignment
        * Detect in the filtered phase the closest peak to the cue. This step
          is repeated to each trial in order to get a list of length (n_epochs)
          that contains the number of sample (shift) so that if the phase is
          moved, the peak fall onto the cue. A positive shift indicates that
          the phase is moved forward while a negative shift is for a backward
          move
        * Apply, to each trial, this shift to the amplitude
        * Plot the mean re-aligned amplitudes

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_epochs, n_times)
    sf : float
        The sampling frequency
    cue : int, float
        Time-point to use in order to detect the closest phase peak. This
        parameter works in conjunction with the `times` input below. Use
        either :

            * An integer and `times` is None to indicate that you want to
              realign according to a time-point in sample
            * A integer or a float with `times` the time vector if you want
              that Tensorpac automatically infer the sample number around which
              to align
    times : array_like | None
        Time vector
    f_pha : tuple, list | [2, 4]
        List of two floats describing the frequency bounds for extracting the
        phase
    f_amp : tuple, list | [60, 80]
        Frequency vector for the amplitude. Here you can use several forms to
        define those vectors :

            * Dynamic definition : (start, stop, width, step)
            * Using a string : `f_amp` can be 'lres', 'mres', 'hres'
              respectively for low, middle and high resolution vectors
    cycle : tuple | (3, 6)
        Control the number of cycles for filtering. Should be a tuple of
        integers where the first one refers to the number of cycles for the
        phase and the second for the amplitude
        :cite:`bahramisharif2013propagating`.
    """

    def __init__(self, x, sf, cue, times=None, f_pha=[5, 7], f_amp='hres',
                 cycle=(3, 6), n_jobs=-1, verbose=None):
        """Init."""
        set_log_level(verbose)
        # initialize to retrieve filtering methods
        _PacObj.__init__(self, f_pha=f_pha, f_amp=f_amp, dcomplex='hilbert',
                         cycle=cycle)
        _PacVisual.__init__(self)
        logger.info("PeakLockedTF object defined")
        # inputs checking
        x = np.atleast_2d(x)
        assert isinstance(x, np.ndarray) and (x.ndim == 2)
        assert isinstance(sf, (int, float))
        assert isinstance(cue, (int, float))
        assert isinstance(f_pha, (list, tuple)) and (len(f_pha) == 2)
        n_epochs, n_times = x.shape

        # manage cur conversion
        if times is None:
            cue = int(cue)
            times = np.arange(n_times)
            logger.info(f"    align on sample cue={cue}")
        else:
            assert isinstance(times, np.ndarray) and (len(times) == n_times)
            cue_time = cue
            cue = np.abs(times - cue).argmin() - 1
            logger.info(f"    align on time-point={cue_time} (sample={cue})")
        self.cue, self._times = cue, times

        # extract phase and amplitudes
        logger.info(f"    extract phase and amplitudes "
                    f"(n_amps={len(self.yvec)})")
        kw = dict(keepfilt=False, n_jobs=n_jobs)
        pha = self.filter(sf, x, 'phase', n_jobs=n_jobs, keepfilt=True)
        amp = self.filter(sf, x, 'amplitude', n_jobs=n_jobs)
        self._pha, self._amp = pha, amp ** 2

        # peak detection
        logger.info(f"    running peak detection around sample={cue}")
        self.shifts = self._peak_detection(self._pha.squeeze(), cue)

        # realign phases and amplitudes
        logger.info(f"    realign the {n_epochs} phases and amplitudes")
        self.amp_a = self._shift_signals(self._amp, self.shifts, fill_with=0.)
        self.pha_a = self._shift_signals(self._pha, self.shifts, fill_with=0.)

    @staticmethod
    def _peak_detection(pha, cue):
        """Single trial closest to a cue peak detection.

        Parameters
        ----------
        pha : array_like
            Array of single trial phases of shape (n_trials, n_times)
        cue : int
            Cue to use as a reference (in sample unit)

        Returns
        -------
        peaks : array_like
            Array of length (n_trials,) describing each delay to apply
            to each trial in order to realign the phases. In detail :

                * Positive delays means that zeros should be prepend
                * Negative delays means that zeros should be append
        """
        n_trials, n_times = pha.shape
        peaks = []
        for tr in range(n_trials):
            # select the single trial phase
            st_pha = pha[tr, :]
            # detect all peaks across time points
            st_peaks = []
            for t in range(n_times - 1):
                if (st_pha[t - 1] < st_pha[t]) and (st_pha[t] > st_pha[t + 1]):
                    st_peaks += [t]
            # detect the minimum peak
            min_peak = st_peaks[np.abs(np.array(st_peaks) - cue).argmin()]
            peaks += [cue - min_peak]

        return np.array(peaks)

    @staticmethod
    def _shift_signals(sig, n_shifts, fill_with=0):
        """Shift an array of signals according to an array of delays.

        Parameters
        ----------
        sig : array_like
            Array of signals of shape (n_freq, n_trials, n_times)
        n_shifts : array_like
            Array of delays to apply to each trial of shape (n_trials,)
        fill_with : int
            Value to prepend / append to each shifted time-series

        Returns
        -------
        sig_shifted : array_like
            Array of shifted signals with the same shape as the input
        """
        # prepare the needed variables
        n_freqs, n_trials, n_pts = sig.shape
        sig_shifted = np.zeros_like(sig)
        # shift each trial
        for tr in range(n_trials):
            # select the data of a specific trial
            st_shift = n_shifts[tr]
            st_sig = sig[:, tr, :]
            fill = np.full((n_freqs, abs(st_shift)), fill_with,
                           dtype=st_sig.dtype)
            # shift this specific trial
            if st_shift > 0:   # move forward = prepend zeros
                sig_shifted[:, tr, :] = np.c_[fill, st_sig][:, 0:-st_shift]
            elif st_shift < 0:  # move backward = append zeros
                sig_shifted[:, tr, :] = np.c_[st_sig, fill][:, abs(st_shift):]

        return sig_shifted

    def plot(self, zscore=False, baseline=None, edges=0, **kwargs):
        """Integrated Peak-Locked TF plotting function.

        Parameters
        ----------
        zscore : bool | False
            Normalize the power by using a z-score normalization. This can be
            useful in order to compensate the 1 / f effect in the power
            spectrum. If True, the mean and deviation are computed at the
            single trial level and across all time points
        baseline : tuple | None
            Baseline period to use in order to apply the z-score correction.
            Should be in samples.
        edges : int | 0
            Number of pixels to discard to compensate filtering edge effect
            (`power[edges:-edges]`).
        kwargs : dict | {}
            Additional arguments are sent to the
            :class:`tensorpac.utils.PeakLockedTF.pacplot` method
        """
        # manage additional arguments
        kwargs['colorbar'] = False
        kwargs['ylabel'] = 'Frequency for amplitude (hz)'
        kwargs['xlabel'] = ''
        kwargs['fz_labels'] = kwargs.get('fz_labels', 14)
        kwargs['fz_cblabel'] = kwargs.get('fz_cblabel', 14)
        kwargs['fz_title'] = kwargs.get('fz_title', 16)
        sl_times = slice(edges, len(self._times) - edges)
        times = self._times[sl_times]
        pha_n = self.pha_a[..., sl_times].squeeze()
        # z-score normalization
        if zscore:
            if baseline is None:
                bsl_idx = sl_times
            else:
                assert len(baseline) == 2
                bsl_idx = slice(baseline[0], baseline[1])
            _mean = self.amp_a[..., bsl_idx].mean(2, keepdims=True)
            _std = self.amp_a[..., bsl_idx].std(2, keepdims=True)
            _std[_std == 0.] = 1.  # correction from NaN
            amp_n = (self.amp_a[..., sl_times] - _mean) / _std
        else:
            amp_n = self.amp_a[..., sl_times]

        # grid definition
        gs = GridSpec(8, 8)
        # image plot
        plt.subplot(gs[slice(0, 6), 0:-1])
        self.pacplot(amp_n.mean(1), times, self.yvec, **kwargs)
        plt.axvline(times[self.cue], color='w', lw=2)
        plt.tick_params(bottom=False, labelbottom=False)
        ax_1 = plt.gca()
        # external colorbar
        plt.subplot(gs[slice(1, 5), -1])
        cb = plt.colorbar(self._plt_im, pad=0.01, cax=plt.gca())
        cb.set_label('Power (V**2/Hz)', fontsize=kwargs['fz_cblabel'])
        cb.outline.set_visible(False)
        # phase plot
        plt.subplot(gs[slice(6, 8), 0:-1])
        plt.plot(times, pha_n.T, color='lightgray', alpha=.2, lw=1.)
        plt.plot(times, pha_n.mean(0), label='single trial phases', alpha=.2,
                 lw=1.)  # legend tweaking
        plt.plot(times, pha_n.mean(0), label='mean phases',
                 color='#1f77b4')
        plt.axvline(times[self.cue], color='k', lw=2)
        plt.autoscale(axis='both', tight=True, enable=True)
        plt.xlabel("Times", fontsize=kwargs['fz_labels'])
        plt.ylabel("V / Hz", fontsize=kwargs['fz_labels'])
        # bottom legend
        plt.legend(loc='center', bbox_to_anchor=(.5, -.5),
                   fontsize='x-large', ncol=2)
        ax_2 = plt.gca()

        return [ax_1, ax_2]



"""Individual methods for assessing PAC."""
import numpy as np
from scipy.special import erfinv

from tensorpac.gcmi import nd_mi_gg


def mean_vector_length(pha, amp):
    """Tensor-based Mean Vector Length (MVL).

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) and
        the array of amplitudes of shape (n_amp, ..., n_times).

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Canolty et al. 2006 :cite:`canolty2006high`
    """
    return np.abs(np.einsum('i...j, k...j->ik...', amp,
                            np.exp(1j * pha))) / pha.shape[-1]


def modulation_index(pha, amp, n_bins=18):
    """Tensor-based Modulation index (MI).

    The modulation index is obtained using the Kullback Leibler Distance which
    measures how much the distribution of binned amplitude differs from a
    uniform distribution.

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) and
        the array of amplitudes of shape (n_amp, ..., n_times).
    n_bins : int | 18
        Number of bins to binarize the amplitude according to phase intervals

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Tort et al. 2010 :cite:`tort2010measuring`
    """
    # Get the phase locked binarized amplitude :
    p_j = _kl_hr(pha, amp, n_bins)
    # Divide the binned amplitude by the mean over the bins :
    p_j /= p_j.sum(axis=0, keepdims=True)
    # Take the log of non-zero values :
    p_j = p_j * np.ma.log(p_j).filled(-np.inf)
    # Compute the PAC :
    pac = 1 + p_j.sum(axis=0) / np.log(n_bins)
    # Set distribution distances that are really closed to zero :
    pac[np.isinf(pac)] = 0.
    return pac


def heights_ratio(pha, amp, n_bins=18):
    """Tensor-based Heights ratio (HR).

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) and
        the array of amplitudes of shape (n_amp, ..., n_times).
    n_bins : int | 18
        Number of bins to binarize the amplitude according to phase intervals

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Lakatos et al. 2005 :cite:`lakatos2005oscillatory`
    """
    # Get the phase locked binarized amplitude :
    p_j = _kl_hr(pha, amp, n_bins)
    # Divide the binned amplitude by the mean over the bins :
    p_j /= p_j.sum(axis=0, keepdims=True)
    # Find (maxximum, minimum) of the binned distribution :
    h_max, h_min = p_j.max(axis=0), p_j.min(axis=0)
    # Compute pac :
    pac = (h_max - h_min) / h_max
    return pac


def _kl_hr(pha, amp, n_bins, mean_bins=True):
    """Binarize the amplitude according to phase values.

    This function is shared by the Kullback-Leibler Distance and the
    Height Ratio.
    """
    vecbin = np.linspace(-np.pi, np.pi, n_bins + 1)
    phad = np.digitize(pha, vecbin) - 1

    abin = []
    for i in np.unique(phad):
        # Find where phase take vecbin values :
        idx = phad == i
        m = idx.sum() if mean_bins else 1.
        # Take the sum of amplitude inside the bin :
        abin_pha = np.einsum('i...j, k...j->ik...', amp, idx) / m
        abin.append(abin_pha)

    return np.array(abin)


def norm_direct_pac(pha, amp, p=.05):
    """Tensor-based Normalized direct Pac (ndPAC).

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) and
        the array of amplitudes of shape (n_amp, ..., n_times).
    p : float | .05
        P-value to use for thresholding. Sub-threshold PAC values
        will be set to 0. To disable this behavior (no masking), use ``p=1`` or
        ``p=None``.

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Ozkurt et al. :cite:`ozkurt2012statistically`
    """
    n_times = amp.shape[-1]
    # normalize amplitude
    # use the sample standard deviation, as in original matlab code from author
    amp = np.subtract(amp, np.mean(amp, axis=-1, keepdims=True))
    amp = np.divide(amp, np.std(amp, ddof=1, axis=-1, keepdims=True))
    # compute pac
    pac = np.abs(np.einsum('i...j, k...j->ik...', amp, np.exp(1j * pha)))

    # no thresholding
    if p == 1. or p is None:
        return pac / n_times

    s = pac ** 2
    pac /= n_times
    # set to zero non-significant values
    xlim = n_times * erfinv(1 - p) ** 2
    pac[s <= 2 * xlim] = 0.
    return pac


def phase_locking_value(pha, pha_amp):
    """Tensor-based Phase Locking-Value (PLV).

    In order to measure the phase locking value, the phase of the amplitude of
    the higher-frequency signal must be provided, and not the amplitude as in
    most other PAC functions.

    Parameters
    ----------
    pha, pha_amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) for
        the lower frequency and the array of phase of the amplitude signal of
        shape (n_pha_amp, ..., n_times) for the higher frequency.

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_pha_amp, n_pha, ...)

    References
    ----------
    Penny et al. 2008 :cite:`penny2008testing`, Lachaux et al. 1999
    :cite:`lachaux1999measuring`
    """
    pac = np.einsum('i...j, k...j->ik...', np.exp(-1j * pha_amp),
                    np.exp(1j * pha))
    return np.abs(pac) / pha.shape[-1]


def gauss_cop_pac(pha, amp):
    """Tensor-based Gaussian Copula PAC (gcPac).

    This function assumes that phases and amplitudes have already been
    prepared i.e. phases should be represented in a unit circle
    (np.c_[np.sin(pha), np.cos(pha)]) and both inputs should also have been
    copnormed.

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) and
        the array of amplitudes of shape (n_amp, ..., n_times).

    Returns
    -------
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)

    References
    ----------
    Ince et al 2017. :cite:`ince2017statistical`
    """
    # prepare the shape of gcpac
    n_pha, n_amp = pha.shape[0], amp.shape[0]
    pha_sh = list(pha.shape[:-2])
    gc = np.zeros([n_amp] + pha_sh, dtype=float)
    # compute mutual information
    for p in range(n_pha):
        for a in range(n_amp):
            gc[a, p, ...] = nd_mi_gg(pha[p, ...], amp[a, ...])
    return gc



"""Test tensorpac functions."""
import numpy as np
from scipy import stats

import matplotlib
import matplotlib.pyplot as plt

from tensorpac import Pac, EventRelatedPac, PreferredPhase
from tensorpac.utils import pac_trivec
from tensorpac.signals import pac_signals_wavelet



def normal(x, mu, sigma):
    return ( 2. * np.pi * sigma ** 2. ) ** -.5 * np.exp(
                -.5 * (x - mu) ** 2. / sigma ** 2. )


class TestPac(object):
    """Tests for Pac, Erpac and Preferred phase."""

    def test_id_pac_definition(self):
        """Test Pac object definition."""
        nmeth, nsuro, nnorm = 6, 4, 5
        for k in range(nmeth):
            for i in range(nsuro):
                for j in range(nnorm):
                    p = Pac(idpac=(k + 1, i, j))
                    str(p)

    def test_filtering_definition(self):
        """Test filtering defintion."""
        dcomplex = ['hilbert', 'wavelet']
        cycle = (12, 24)
        width = 12
        for k in dcomplex:
            Pac(dcomplex=k, cycle=cycle, width=width)

    def test_spectral(self):
        """Test filtering using the provided filters."""
        data = np.random.rand(3, 1000)
        p = Pac()
        dcomplex = ['hilbert', 'wavelet']
        for k in dcomplex:
            p.dcomplex = k
            p.filter(1024, data, n_jobs=1)

    def test_filter(self):
        """Test filter method."""
        data = np.random.rand(2, 1000)
        p = Pac()
        p.filter(256, data, 'phase')
        p.filter(256, data, 'phase', edges=2)
        p.filter(256, data, 'amplitude')

    def test_fit(self):
        """Test all Pac methods."""
        pha = np.random.rand(2, 7, 1024)
        amp = np.random.rand(3, 7, 1024)
        nmeth, nsuro, nnorm = 5, 4, 5
        p = Pac(verbose=False)
        for k in range(nmeth):
            for i in range(nsuro):
                for j in range(nnorm):
                    p.idpac = (k + 1, i, j)
                    p.fit(pha, amp, n_jobs=1, n_perm=10)
                    if (i >= 1) and (k + 1 != 4):
                        for mcp in ['maxstat', 'fdr', 'bonferroni']:
                            p.infer_pvalues(mcp=mcp)

    def test_filterfit(self):
        """Test filtering test computing PAC."""
        data = np.random.rand(2, 1024)
        p = Pac(idpac=(4, 0, 0))
        p.filterfit(1024, data, n_jobs=1)
        p.idpac = (1, 1, 1)
        p.filterfit(1024, data, data, n_jobs=1, n_perm=2)
        p.dcomplex = 'wavelet'
        p.filter(1024, data, n_jobs=1)

    def test_properties(self):
        """Test Pac properties."""
        p = Pac()
        # Idpac :
        p.idpac
        p.idpac = (2, 1, 1)
        # Dcomplex :
        p.dcomplex
        p.dcomplex = 'wavelet'
        # Cycle :
        p.cycle
        p.cycle = (12, 24)
        # Width :
        p.width
        p.width = 12

    def test_pac_comodulogram(self):
        """Test Pac object definition.

        This test works locally but failed on travis...
        """
        matplotlib.use('agg')
        f, tridx = pac_trivec()
        pac = np.random.rand(20, 10)
        pval = np.random.rand(20, 10)
        p = Pac(f_pha=np.arange(11), f_amp=np.arange(21))
        p.comodulogram(np.random.rand(10, 10, 20))
        p.comodulogram(pac, rmaxis=True, dpaxis=True, interp=(.1, .1))
        p.comodulogram(pac, plotas='contour', pvalues=pval)
        p.comodulogram(pac, plotas='pcolor', pvalues=pval, levels=[.5, .7],
                       under='gray', over='red', bad='orange')
        p = Pac(f_pha=np.arange(11), f_amp=f)
        pac = np.random.rand(len(f))
        p.triplot(pac, f, tridx)
        p.savefig('test_savefig.png')
        p.show()
        matplotlib.pyplot.close('all')

    def test_functional_pac(self):
        """Test functionnal pac."""
        # generate a 10<->100hz ground truth coupling
        n_epochs, sf, n_times = 1, 512., 4000
        data, time = pac_signals_wavelet(f_pha=10, f_amp=100, noise=.8,
                                         n_epochs=n_epochs, n_times=n_times,
                                         sf=sf)
        # phase / amplitude extraction (single time)
        p = Pac(f_pha='lres', f_amp='lres', dcomplex='wavelet', width=12)
        phases = p.filter(sf, data, ftype='phase', n_jobs=1)
        amplitudes = p.filter(sf, data, ftype='amplitude', n_jobs=1)
        # ground truth array construction
        n_pha, n_amp = len(p.xvec), len(p.yvec)
        n_pix = int(n_pha * n_amp)
        gt = np.zeros((n_amp, n_pha), dtype=bool)
        b_pha = np.abs(p.xvec.reshape(-1, 1) - np.array([[9, 11]])).argmin(0)
        b_amp = np.abs(p.yvec.reshape(-1, 1) - np.array([[95, 105]])).argmin(0)
        gt[b_amp[0]:b_amp[1] + 1, b_pha[0]:b_pha[1] + 1] = True

        plt.figure(figsize=(12, 9))
        plt.subplot(2, 3, 1)
        p.comodulogram(gt, title='Gound truth', cmap='magma', colorbar=False)
        # loop over implemented methods
        for i, k in enumerate([1, 2, 3, 5, 6]):
            # compute only the pac
            p.idpac = (k, 2, 3)
            xpac = p.fit(phases, amplitudes, n_perm=200).squeeze()
            pval = p.pvalues.squeeze()
            is_coupling = pval <= .05
            # count the number of correct pixels. This includes both true
            # positives and true negatives
            acc = 100 * (is_coupling == gt).sum() / n_pix
            assert acc > 95.
            # build title of the figure (for sanity check)
            meth = p.method.replace(' (', '\n(')
            title = f"Method={meth}\nAccuracy={np.around(acc, 2)}%"
            # set to nan everywhere it's not significant
            xpac[~is_coupling] = np.nan
            vmin, vmax = np.nanmin(xpac), np.nanmax(xpac)
            # plot the results
            plt.subplot(2, 3, i + 2)
            p.comodulogram(xpac, colorbar=False, vmin=vmin, vmax=vmax,
                           title=title)
            plt.ylabel(''), plt.xlabel('')
        plt.tight_layout()
        plt.show()  # show on demand


class TestErpac(object):
    """Test EventRelatedPac class."""

    def test_filter(self):
        """Test function filter."""
        data = np.random.rand(7, 1000)
        p = EventRelatedPac()
        p.filter(256, data, 'phase')
        p.filter(256, data, 'amplitude')

    def test_fit(self):
        """Test function fit."""
        data = np.random.rand(100, 1000)
        p = EventRelatedPac()
        pha = p.filter(256, data, 'phase')
        amp = p.filter(256, data, 'amplitude')
        p.fit(pha, amp, method='circular')
        p.fit(pha, amp, method='gc')
        p.fit(pha, amp, method='gc', n_perm=2)
        p.fit(pha, amp, method='gc', smooth=5)
        p.surrogates, p.pvalues

    def test_filterfit(self):
        """Test function filterfit."""
        p = EventRelatedPac()
        x_pha = np.random.rand(100, 1000)
        x_amp = np.random.rand(100, 1000)
        p.filterfit(256, x_pha, x_amp=x_amp, method='circular')
        p.filterfit(256, x_pha, x_amp=x_amp, method='gc')

    def test_functional_erpac(self):
        """Test function test_functional_pac."""
        # erpac simultation
        n_epochs, n_times, sf, edges = 400, 1000, 512., 50
        x, times = pac_signals_wavelet(f_pha=10, f_amp=100, n_epochs=n_epochs,
                                       noise=.1, n_times=n_times, sf=sf)
        times = times[edges:-edges]
        # phase / amplitude extraction (single time)
        p = EventRelatedPac(f_pha=[8, 12], f_amp=(30, 200, 5, 5),
                            dcomplex='wavelet', width=12)
        kw = dict(n_jobs=1, edges=edges)
        phases = p.filter(sf, x, ftype='phase', **kw)
        amplitudes = p.filter(sf, x, ftype='amplitude', **kw)
        n_amp = len(p.yvec)
        # generate a normal distribution
        gt = np.zeros((n_amp, n_times - 2 * edges))
        b_amp = np.abs(p.yvec.reshape(-1, 1) - np.array([[80, 120]])).argmin(0)
        gt[b_amp[0]:b_amp[1] + 1, :] = True

        plt.figure(figsize=(16, 5))
        plt.subplot(131)
        p.pacplot(gt, times, p.yvec, title='Ground truth', cmap='magma')

        for n_meth, meth in enumerate(['circular', 'gc']):
            # compute erpac + p-values
            erpac = p.fit(phases, amplitudes, method=meth,
                          mcp='bonferroni', n_perm=30).squeeze()
            pvalues = p.pvalues.squeeze()
            # find everywhere erpac is significant + compare to ground truth
            is_signi = pvalues < .05
            erpac[~is_signi] = np.nan
            # computes accuracy
            acc = 100 * (is_signi == gt).sum() / (n_amp * n_times)
            assert acc > 80.
            # plot the result
            title = f"Method={p.method}\nAccuracy={np.around(acc, 2)}%"
            plt.subplot(1, 3, n_meth + 2)
            p.pacplot(erpac, times, p.yvec, title=title)
        plt.tight_layout()
        plt.show()



class TestPreferredPhase(object):
    """Test EventRelatedPac class."""

    def test_filter(self):
        """Test function filter."""
        data = np.random.rand(7, 1000)
        p = PreferredPhase()
        p.filter(256, data, 'phase')
        p.filter(256, data, 'amplitude')

    def test_fit(self):
        """Test function fit."""
        data = np.random.rand(100, 1000)
        p = PreferredPhase()
        pha = p.filter(256, data, 'phase')
        amp = p.filter(256, data, 'amplitude')
        p.fit(pha, amp)

    def test_filterfit(self):
        """Test function filterfit."""
        p = PreferredPhase()
        x_pha = np.random.rand(100, 1000)
        x_amp = np.random.rand(100, 1000)
        p.filterfit(256, x_pha, x_amp=x_amp)

    def test_polar_plot(self):
        """Test the polar plot."""
        matplotlib.use('agg')
        p = PreferredPhase(f_pha=[5, 7], f_amp=(60, 200, 10, 1))
        x_pha = np.random.rand(100, 1000)
        x_amp = np.random.rand(100, 1000)
        ampbin, pp, vecbin = p.filterfit(256, x_pha, x_amp=x_amp)
        pp = np.squeeze(pp).T
        ampbin = np.squeeze(ampbin).mean(-1)
        p.polar(ampbin.T, vecbin, p.yvec, cmap='RdBu_r', interp=.1,
                cblabel='Amplitude bins')
        matplotlib.pyplot.close('all')


if __name__ == '__main__':
    TestPac().test_fit()
    # TestErpac().test_functional_erpac()


"""Individual methods for assessing surrogates."""
import numpy as np
from joblib import Parallel, delayed

from tensorpac.config import CONFIG


def compute_surrogates(pha, amp, ids, fcn, n_perm, n_jobs, random_state):
    """Compute surrogates using tensors and parallel computing."""
    if ids == 0:
        return None
    else:
        fcn_p = {1: swap_pha_amp, 2: swap_blocks, 3: time_lag}[ids]

    def para_surr(rnd):  # noqa
        return fcn(*fcn_p(pha, amp, random_state=random_state + rnd))
    s = Parallel(n_jobs=n_jobs, **CONFIG['JOBLIB_CFG'])(delayed(para_surr)(
        k) for k in range(n_perm))
    return np.array(s)


def swap_pha_amp(pha, amp, random_state=None):
    """Compute surrogates by swapping phase / amplitude trials.

    This function destroys the relation between the phase and the amplitude in
    order to correct for PAC that could be obtained by chance. To this end,
    this function reorganize the relation from trial-to-trial between the phase
    and the amplitude (swapping). In particular, in that case it is the trials
    of the phase that are randomly swapped and the amplitude is leaved
    unchanged. For a more detailed description, see :cite:`tort2010measuring`

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape
        (n_pha, n_trials, ..., n_times) and the array of amplitudes of shape
        (n_amp, n_trials, ..., n_times).
    random_state : int | None
        Fix the random state of the machine for reproducible results.

    Returns
    -------
    pha, amp : array_like
        The phase and amplitude to use to compute the distribution of
        permutations

    References
    ----------
    Tort et al. 2010 :cite:`tort2010measuring`
    """
    if random_state is None:
        random_state = int(np.random.randint(0, 10000, size=1))
    rnd = np.random.RandomState(random_state)
    tr_ = rnd.permutation(pha.shape[1])
    return pha[:, tr_, ...], amp


def swap_blocks(pha, amp, random_state=None):
    """Compute surrogates by swapping amplitudes time blocks.

    This function destroys the relation between the phase and the amplitude in
    order to correct for PAC that could be obtained by chance. To this end,
    this function cut the amplitude at a random time point. Then, both time
    blocks are swapped. For a more detailed description, see
    :cite:`bahramisharif2013propagating`.

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) and
        the array of amplitudes of shape (n_amp, ..., n_times).
    random_state : int | None
        Fix the random state of the machine for reproducible results.

    Returns
    -------
    pha, amp : array_like
        The phase and amplitude to use to compute the distribution of
        permutations

    References
    ----------
    Bahramisharif et al. 2013 :cite:`bahramisharif2013propagating`
    """
    if random_state is None:
        random_state = int(np.random.randint(0, 10000, size=1))
    rnd = np.random.RandomState(random_state)
    # random cutting point along time axis
    cut_at = rnd.randint(1, amp.shape[-1], (1,))
    # Split amplitude across time into two parts :
    ampl = np.array_split(amp, cut_at, axis=-1)
    # Revered elements :
    ampl.reverse()
    return pha, np.concatenate(ampl, axis=-1)


def time_lag(pha, amp, random_state=None):
    """Compute surrogates by introducing a time lag on phase series.

    This function destroys the relation between the phase and the amplitude in
    order to correct for PAC that could be obtained by chance. To this end, a
    random temporal lag is introduced at the beginning of the phase. For a more
    detailed description, see :cite:`canolty2006high`.

    Parameters
    ----------
    pha, amp : array_like
        Respectively the arrays of phases of shape (n_pha, ..., n_times) and
        the array of amplitudes of shape (n_amp, ..., n_times).
    random_state : int | None
        Fix the random state of the machine for reproducible results.

    Returns
    -------
    pha, amp : array_like
        The phase and amplitude to use to compute the distribution of
        permutations

    References
    ----------
    Canolty et al. 2006 :cite:`canolty2006high`
    """
    if random_state is None:
        random_state = int(np.random.randint(0, 10000, size=1))
    rnd = np.random.RandomState(random_state)
    shift = rnd.randint(pha.shape[-1])
    return np.roll(pha, shift, axis=-1), amp


def normalize(idn, pac, surro):
    """Normalize the phase amplitude coupling.

    This function performs inplace normalization (i.e. without copy of array)

    Parameters
    ----------
    idn : int
        Normalization method to use :

            * 1 : substract the mean of surrogates
            * 2 : divide by the mean of surrogates
            * 3 : substract then divide by the mean of surrogates
            * 4 : substract the mean then divide by the deviation of surrogates
    pac : array_like
        Array of phase amplitude coupling of shape (n_amp, n_pha, ...)
    surro : array_like
        Array of surrogates of shape (n_perm, n_amp, n_pha, ...)
    """
    s_mean, s_std = np.mean(surro, axis=0), np.std(surro, axis=0)
    if idn == 1:  # Substraction
        pac -= s_mean
    elif idn == 2:  # Divide
        pac /= s_mean
    elif idn == 3:  # Substract then divide
        pac -= s_mean
        pac /= s_mean
    elif idn == 4:  # Z-score
        pac -= s_mean
        pac /= s_std



"""Statistics for cross frequency coupling."""
import logging

import numpy as np

from tensorpac.io import is_pandas_installed, is_statsmodels_installed

logger = logging.getLogger('tensorpac')


def test_stationarity(x, p=.05):
    """Test the stationarity of an electrophysiological dataset.

    This function performs a Augmented Dickey-Fuller test and returns a table
    (dataframe) with statistical properties for each epoch.

    Parameters
    ----------
    x : array_like
        Array of data of shape (n_epochs, n_times)
    p : float | 0.05
        P-value to use as a threshold in order to infer if the time-series
        are significantly stationary

    Returns
    -------
    df : pandas.DataFrame
        Dataframe that contains the statistical properties for each
        epoch. The table contains the followong columns :

            * Epochs : epoch number
            * P-values
            * Stationary : boolean indicating if the time-serie is
              significantly stationary at the critical level p
            * Statistics : statistical test
            * CV (5%) and CV (1%) : critical value respectively at 5% and 1%

    Notes
    -----
    This function requires both pandas and statsmodels Python packages. See :

        * Pandas : https://pandas.pydata.org/pandas-docs/stable/install.html
        * Statsmodels : http://www.statsmodels.org/stable/install.html
    """
    is_pandas_installed()
    is_statsmodels_installed()
    import pandas as pd
    from statsmodels.tsa.stattools import adfuller
    x = np.atleast_2d(x)
    assert x.ndim == 2, ("x should be a two dimentional array of shape "
                         "(n_epochs, n_times)")
    n_epochs, n_times = x.shape
    # test time series
    logger.info(f"Performing a Augmented Dickey-Fuller test on {n_epochs} "
                f"epochs with p={p}")
    epochs = ['epoch %i' % k for k in range(n_epochs)]
    pvalues = np.zeros((n_epochs,), dtype=float)
    adf_stat = np.zeros_like(pvalues)
    stationary = np.zeros_like(pvalues, dtype=bool)
    cv_5, cv_1 = np.zeros_like(pvalues), np.zeros_like(pvalues)
    for k in range(n_epochs):
        result = adfuller(x[k, :])
        adf_stat[k], pvalues[k] = result[0], result[1]
        cv_5[k], cv_1[k] = result[4]['5%'], result[4]['1%']
        stationary[k] = result[1] <= p
    # pandas formating
    cols = ['Epochs', 'P-values', f'Stationary', 'Statistics',
            'CV (5%)', 'CV (1%)']
    df = pd.DataFrame({cols[0]: epochs, cols[1]: pvalues, cols[2]: stationary,
                       cols[3]: adf_stat, cols[4]: cv_5, cols[5]: cv_1},
                      columns=cols)
    n_signi = stationary.sum()
    # n_not_signi = (~stationary.sum())
    logger.info(f"    {n_signi}/{n_epochs} epochs were found as significantly "
                f"stationary at p={p}")
    return df



"""Individual methods for assessing Preferred Phase."""
import numpy as np

from .meth_pac import _kl_hr


def preferred_phase(pha, amp, n_bins=18):
    """Compute the preferred phase.

    Parameters
    ----------
    pha, amp : array_like
        Respectively the phase of slower oscillations of shape
        (n_pha, n_epochs, n_times) and the amplitude of faster
        oscillations of shape (n_pha, n_epochs, n_times).
    n_bins : int | 72
        Number of bins for bining the amplitude according to phase
        slices.

    Returns
    -------
    binned_amp : array_like
        The binned amplitude according to the phase of shape
        (n_bins, n_amp, n_pha, n_epochs).
    pp : array_like
        The prefered phase where the amplitude is maximum of shape
        (namp, npha, n_epochs).
    polarvec : array_like
        The phase vector for the polar plot of shape (n_bins,)
    """
    # Bin the amplitude according to the phase :
    binned_amp = _kl_hr(pha, amp, n_bins)
    binned_amp /= binned_amp.sum(axis=0, keepdims=True)
    # Find the index where the amplitude is maximum over the bins :
    idxmax = binned_amp.argmax(axis=0)
    # Find the preferred phase :
    binsize = (2 * np.pi) / float(n_bins)
    vecbin = np.arange(-np.pi, np.pi, binsize) + binsize / 2
    pp = vecbin[idxmax]
    # Build the phase vector (polar plot) :
    polarvec = np.linspace(-np.pi, np.pi, binned_amp.shape[0])
    return binned_amp, pp, polarvec



"""Test tensorpac utils."""
import numpy as np
import matplotlib

from tensorpac.utils import (pac_vec, pac_trivec, PSD, BinAmplitude, ITC,
                             PeakLockedTF)


class TestUtils(object):
    """Test utility functions."""

    def test_pac_vec(self):
        """Definition of PAC vectors."""
        assert pac_vec()
        assert pac_vec(f_pha=(1, 30, 2, 2), f_amp=(60, 200, 10, 5))
        assert pac_vec(f_pha=[1, 2], f_amp=np.arange(50))
        assert pac_vec(f_pha=np.array([[2, 4], [5, 7], [9, 10]]),
                       f_amp=np.array([[30, 60], [60, 90], [100, 200]]).T)
        assert pac_vec(f_pha=[[1, 2], [5, 7]], f_amp=[60, 150])
        assert pac_vec(f_pha='lres', f_amp='lres')
        assert pac_vec(f_pha='mres', f_amp='mres')
        assert pac_vec(f_pha='hres', f_amp='hres')

    def test_trivec(self):
        """Definition of triangular vectors."""
        assert pac_trivec(2, 200, 10)

    def test_psd(self):
        """Test PSD."""
        # test definition
        x = np.random.rand(10, 200)
        psd = PSD(x, 128)
        # test properties
        psd.freqs
        psd.psd
        # test plotting
        matplotlib.use('agg')
        psd.plot(confidence=None, log=True, grid=True, interp=.1)
        psd.plot(confidence=.95, log=False, grid=False, interp=None)
        psd.plot_st_psd(log=True, grid=True)
        psd.show()
        matplotlib.pyplot.close('all')

    def test_binned_amplitude(self):
        """Test binned amplitude."""
        # test definition
        x = np.random.rand(10, 200)
        binamp = BinAmplitude(x, 128)
        # test plot
        binamp.plot(unit='rad')
        binamp.plot(unit='deg')
        binamp.show()
        matplotlib.pyplot.close('all')
        # test properties
        binamp.phase
        binamp.amplitude

    def test_itc(self):
        """Test Phase-locking Value."""
        # test definition
        x = np.random.rand(10, 200)
        itc_1d = ITC(x, 128, f_pha=[2, 4])
        itc_2d = ITC(x, 128, f_pha=[[2, 4], [5, 7]])
        # test properties
        itc_1d.itc
        # test plot
        itc_1d.plot()
        itc_2d.plot()
        itc_1d.show()
        matplotlib.pyplot.close('all')

    def test_peaklockedtf(self):
        """Test class PeakLockedTF."""
        x = np.random.rand(10, 1000)
        times = np.linspace(-1, 1, 1000)
        PeakLockedTF(x, 512, 100,)
        p_obj = PeakLockedTF(x, 512, 0., times=times)
        p_obj.plot(zscore=False)
        p_obj.plot(zscore=True, edges=10)



"""Test PAC methods."""
import numpy as np
from tensorpac.methods.meth_switch import get_pac_fcn, pacstr
from tensorpac.methods.meth_surrogates import compute_surrogates, normalize
from tensorpac.methods.meth_pp import preferred_phase
from tensorpac.methods.meth_erpac import erpac, ergcpac, _ergcpac_perm

n_pac_range = range(1, 7)
n_sur_range = range(4)
n_norm_range = range(1, 5)

n_epochs = 5
n_times = 1000
n_pha_freqs = 2
n_amp_freqs = 3
n_bins = 18
pval = .05
n_perm = 2

pha = np.random.uniform(-np.pi, np.pi, (n_pha_freqs, n_epochs, n_times))
amp = np.random.rand(n_amp_freqs, n_epochs, n_times)


class TestMethods(object):
    """Test individual pac methods."""

    @staticmethod
    def _get_methods(implementation='tensor'):
        meths = []
        q = 1
        while q is not None:
            try:
                meths += [get_pac_fcn(q, n_bins, pval, implementation)]
                q += 1
            except KeyError as e:  # noqa
                q = None
        return meths

    def test_pacstr(self):
        """Test getting pac string name."""
        for p in n_pac_range:
            for s in n_sur_range:
                for n in n_norm_range:
                    names = pacstr((p, s, n))
                    assert all([isinstance(k, str) for k in names])
        # test assert error
        try:
            pacstr((p + 1, s, n))
        except ValueError as e:  # noqa
            pass
        try:
            pacstr((p, s + 1, n))
        except ValueError as e:  # noqa
            pass
        try:
            pacstr((p, s, n + 1))
        except ValueError as e:  # noqa
            pass

    def test_pac_methods(self):
        """Test individual pac methods."""
        for imp in ['tensor']:  # 'numba' = FAIL (core dumped)
            for n, meth in enumerate(self._get_methods(imp)):
                # print(meth.func.__name__)
                if n + 1 == 6:  # gc pac need additional multivariate axis
                    _pha = np.stack([np.sin(pha), np.cos(pha)], axis=-2)
                    _amp = amp[..., np.newaxis, :]
                    pac = meth(_pha, _amp)
                elif n + 1 == 4:  # Try with different values of p for coverage
                    pac = meth(pha, amp, p=0.5)
                    pac = meth(pha, amp, p=1)
                    pac = meth(pha, amp, p=None)
                else:
                    pac = meth(pha, amp)
                assert pac.shape == (n_amp_freqs, n_pha_freqs, n_epochs)

    def test_surrogates(self):
        """Test computing surrogates."""
        fcn = get_pac_fcn(1, n_bins, pval)
        s_shape = (n_perm, n_amp_freqs, n_pha_freqs, n_epochs)
        for s in n_sur_range:
            surro = compute_surrogates(pha, amp, s, fcn, n_perm, 1, 0)
            assert (surro is None) or (surro.shape == s_shape)

    def test_normalize(self):
        """Test normalization."""
        for k in n_norm_range:
            true_pac = np.random.rand(n_amp_freqs, n_pha_freqs)
            perm_pac = np.random.rand(n_perm, n_amp_freqs, n_pha_freqs)
            normalize(k, true_pac, perm_pac)

    def test_erpac(self):
        """Test event-related PAC."""
        er_pha, er_amp = np.moveaxis(pha, -2, -1), np.moveaxis(amp, -2, -1)
        # circular
        er_circ, pv_circ = erpac(er_pha, er_amp)
        assert er_circ.shape == pv_circ.shape
        assert er_circ.shape == (n_amp_freqs, n_pha_freqs, n_times)
        # gaussian copula
        _pha = np.stack([np.sin(er_pha), np.cos(er_pha)], axis=-2)
        _amp = er_amp[..., np.newaxis, :]
        ergc_circ = ergcpac(_pha, _amp, smooth=None)
        assert ergc_circ.shape == (n_amp_freqs, n_pha_freqs, n_times)
        ergcpac(_pha, _amp, smooth=5)
        # test erpac permutations
        ergc_perm = _ergcpac_perm(_pha, _amp, smooth=None, n_perm=n_perm)
        assert ergc_perm.shape == (n_perm, n_amp_freqs, n_pha_freqs, n_times)

    def test_preferred_phase(self):
        """Test preferred phase method."""
        bin_amp, pp, fvec = preferred_phase(pha, amp, n_bins)
        assert bin_amp.shape == (n_bins, n_amp_freqs, n_pha_freqs, n_epochs)
        assert pp.shape == (n_amp_freqs, n_pha_freqs, n_epochs)
        assert fvec.shape == (n_bins,)



"""Define tensorpac logger.

See :
https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
"""
import logging
import sys
import re

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"
COLORS = {
    'DEBUG': GREEN,
    'PROFILER': MAGENTA,
    'INFO': WHITE,
    'WARNING': YELLOW,
    'ERROR': RED,
    'CRITICAL': RED,
}
FORMAT = {'compact': "$BOLD%(levelname)s | %(message)s",
          'spacy': "$BOLD%(levelname)-19s$RESET | %(message)s",
          'tensorpac': "$BOLD%(name)s-%(levelname)-19s$RESET | %(message)s",
          'print': "%(message)s",
          }


def formatter_message(message, use_color=True):
    """Format the message."""
    return message.replace("$RESET", RESET_SEQ).replace("$BOLD", BOLD_SEQ)


class _Formatter(logging.Formatter):
    """Formatter."""

    def __init__(self, format_type='compact'):
        logging.Formatter.__init__(self, FORMAT[format_type])
        self._format_type = format_type

    def format(self, record):
        name = record.levelname
        msg = record.getMessage()
        # If * in msg, set it in RED :
        if '*' in msg:
            regexp = '\*.*?\*'
            re_search = re.search(regexp, msg).group()
            to_color = COLOR_SEQ % (30 + RED) + re_search + COLOR_SEQ % (
                30 + WHITE) + RESET_SEQ
            msg_color = re.sub(regexp, to_color, msg)
            msg_color += RESET_SEQ
            record.msg = msg_color
        # Set level color :
        levelname_color = COLOR_SEQ % (30 + COLORS[name]) + name + RESET_SEQ
        record.levelname = levelname_color
        if record.levelno == 20:
            logging.Formatter.__init__(self, FORMAT['print'])
        else:
            logging.Formatter.__init__(self, FORMAT[self._format_type])
        return formatter_message(logging.Formatter.format(self, record))


class _StreamHandler(logging.StreamHandler):
    """Stream handler allowing matching and recording."""

    def __init__(self):
        logging.StreamHandler.__init__(self, sys.stderr)
        self.setFormatter(_lf)
        self._str_pattern = None
        self.emit = self._tensorpac_emit

    def _tensorpac_emit(self, record, *args):
        msg = record.getMessage()
        test = self._match_pattern(record, msg)
        if test:
            record.msg = test
            return logging.StreamHandler.emit(self, record)
        else:
            return

    def _match_pattern(self, record, message):
        if isinstance(self._str_pattern, str):
            if re.search(self._str_pattern, message):
                sub = '*{}*'.format(self._str_pattern)
                return re.sub(self._str_pattern, sub, message)
            else:
                return ''
        else:
            return message


logger = logging.getLogger('tensorpac')
_lf = _Formatter()
_lh = _StreamHandler()  # needs _lf to exist
logger.addHandler(_lh)
PROFILER_LEVEL_NUM = 1
logging.addLevelName(PROFILER_LEVEL_NUM, "PROFILER")


def profiler_fcn(self, message, *args, **kws):
    # Yes, logger takes its '*args' as 'args'.
    if self.isEnabledFor(PROFILER_LEVEL_NUM):
        self._log(PROFILER_LEVEL_NUM, message, args, **kws)


logging.Logger.profiler = profiler_fcn
LOGGING_TYPES = dict(DEBUG=logging.DEBUG, INFO=logging.INFO,
                     WARNING=logging.WARNING, ERROR=logging.ERROR,
                     CRITICAL=logging.CRITICAL, PROFILER=PROFILER_LEVEL_NUM)


def set_log_level(verbose=None, match=None):
    """Convenience function for setting the logging level.

    This function comes from the PySurfer package. See :
    https://github.com/nipy/PySurfer/blob/master/surfer/utils.py

    Parameters
    ----------
    verbose : bool, str, int, or None
        The verbosity of messages to print. If a str, it can be either
        PROFILER, DEBUG, INFO, WARNING, ERROR, or CRITICAL.
    match : string | None
        Filter logs using a string pattern.
    """
    # if verbose is None:
    #     verbose = "INFO"
    logger = logging.getLogger('tensorpac')
    if isinstance(verbose, bool):
        verbose = 'INFO' if verbose else 'WARNING'
    if isinstance(verbose, str):
        if (verbose.upper() in LOGGING_TYPES):
            verbose = verbose.upper()
            verbose = LOGGING_TYPES[verbose]
            logger.setLevel(verbose)
        else:
            raise ValueError("verbose must be in "
                             "%s" % ', '.join(LOGGING_TYPES))
    if isinstance(match, str):
        _lh._str_pattern = match


def progress_bar(value, endvalue, bar_length=20, pre_st=None):
    """Progress bar."""
    percent = float(value) / endvalue
    arrow = '-' * int(round(percent * bar_length) - 1) + '>'
    spaces = ' ' * (bar_length - len(arrow))
    pre_st = '' if not isinstance(pre_st, str) else pre_st

    sys.stdout.write("\r{0} [{1}] {2}%".format(pre_st, arrow + spaces,
                                               int(round(percent * 100))))
    sys.stdout.flush()


def is_pandas_installed():
    """Test if pandas is installed."""
    try:
        import pandas  # noqa
    except:
        raise IOError("pandas not installed. See https://pandas.pydata.org/"
                      "pandas-docs/stable/install.html")


def is_statsmodels_installed():
    """Test if statsmodels is installed."""
    try:
        import statsmodels  # noqa
    except:
        raise IOError("statsmodels not installed. See http://www.statsmodels."
                      "org/stable/install.html")



"""Test stastical functions."""
import numpy as np

from tensorpac.stats import test_stationarity as stats_stationarity

n_epochs = 8
n_times = 200
sf = 128.
pval = .05
rng = np.random.RandomState(1)
data = rng.rand(n_epochs, n_times)


class TestStats(object):
    """Test statistical functions."""

    def test_stats_stationarity(self):
        """Test the stationarity."""
        stats_stationarity(data, p=pval)



"""Main PAC class."""
import numpy as np
import logging

from tensorpac.spectral import spectral, hilbertm
from tensorpac.methods import (get_pac_fcn, pacstr, compute_surrogates,
                               erpac, ergcpac, _ergcpac_perm, preferred_phase,
                               normalize)
from tensorpac.gcmi import copnorm
from tensorpac.visu import _PacVisual, _PacPlt, _PolarPlt
from tensorpac.io import set_log_level
from tensorpac.config import CONFIG

logger = logging.getLogger('tensorpac')


class _PacObj(object):
    """Main class for relative PAC objects."""

    def __init__(self, f_pha=[2, 4], f_amp=[60, 200], dcomplex='hilbert',
                 cycle=(3, 6), width=7):
        # Frequency checking :
        from tensorpac.utils import pac_vec
        self._f_pha, self._f_amp = pac_vec(f_pha, f_amp)
        self._xvec, self._yvec = self.f_pha.mean(1), self.f_amp.mean(1)
        # Check spectral properties :
        self._speccheck(dcomplex, cycle, width)

    def __str__(self):
        """String representation."""
        return self.method

    def filter(self, sf, x, ftype='phase', keepfilt=False, edges=None,
               n_jobs=-1):
        """Filt the data in the specified frequency bands.

        Parameters
        ----------
        sf : float
            The sampling frequency.
        x : array_like
            Array of data of shape (n_epochs, n_times)
        ftype : {'phase', 'amplitude'}
            Specify if you want to extract phase ('phase') or the amplitude
            ('amplitude').
        n_jobs : int | -1
            Number of jobs to compute PAC in parallel. For very large data,
            set this parameter to 1 in order to prevent large memory usage.
        keepfilt : bool | False
            Specify if you only want the filtered data (True). This parameter
            is only available with dcomplex='hilbert' and not wavelet.
        edges : int | None
            Number of samples to discard to avoid edge effects due to filtering

        Returns
        -------
        xfilt : array_like
            The filtered data of shape (n_freqs, n_epochs, n_times)
        """
        # ---------------------------------------------------------------------
        # check inputs
        assert isinstance(sf, (int, float)), ("The sampling frequency must be "
                                              "a float number.")
        # Compatibility between keepfilt and wavelet :
        if (keepfilt is True) and (self._dcomplex is 'wavelet'):
            raise ValueError("Using wavelet for the complex decomposition do "
                             "not allow to get filtered data only. Set the "
                             "keepfilt parameter to False or set dcomplex to "
                             "'hilbert'.")
        assert ftype in ['phase', 'amplitude'], ("ftype must either be 'phase'"
                                                 " or 'amplitude.'")
        mne_epochs_type = CONFIG['MNE_EPOCHS_TYPE']
        if not isinstance(x, np.ndarray) and type(x) in mne_epochs_type:
            x = x.get_data()
            sf = x.info['sfreq']
        if x.ndim == 1:
            x = x[np.newaxis, :]
        assert x.ndim == 2, ("x should be a 2d array like (n_epochs, n_times)")
        # check edges
        if not isinstance(edges, int):
            edges = slice(None)
        else:
            logger.debug(f"    Edges {edges} time samples ignored")
            edges = slice(edges, -edges)
        self._edges = edges

        # ---------------------------------------------------------------------
        # Switch between phase or amplitude :
        if ftype is 'phase':
            tosend = 'pha' if not keepfilt else None
            xfilt = spectral(x, sf, self.f_pha, tosend, self._dcomplex,
                             self._cycle[0], self._width, n_jobs)
        elif ftype is 'amplitude':
            tosend = 'amp' if not keepfilt else None
            xfilt = spectral(x, sf, self.f_amp, tosend, self._dcomplex,
                             self._cycle[1], self._width, n_jobs)
        return xfilt[..., edges]

    def _speccheck(self, dcomplex=None, cycle=None, width=None):
        """Check spectral parameters."""
        # Check cycle :
        if cycle is not None:
            cycle = np.asarray(cycle)
            if (len(cycle) is not 2) or not cycle.dtype == int:
                raise ValueError("Cycle must be a tuple of two integers.")
            else:
                self._cycle = cycle
        # Check complex decomposition :
        if dcomplex is not None:
            if dcomplex not in ['hilbert', 'wavelet']:
                raise ValueError("dcomplex must either be 'hilbert' or "
                                 "'wavelet'.")
            else:
                self._dcomplex = dcomplex
        # Convert Morlet's width :
        if width is not None:
            self._width = int(width)

    @staticmethod
    def _phampcheck(pha, amp):
        """Check phase and amplitude values."""
        assert pha.ndim == 3, ("`pha` should have a shape of (n_pha, n_epochs,"
                               " n_times)")
        assert amp.ndim == 3, ("`amp` should have a shape of (n_pha, n_epochs,"
                               " n_times)")
        assert pha.shape[1:] == amp.shape[1:], ("`pha` and `amp` must have the"
                                                " same number of trials, "
                                                "channels and time points")
        assert np.ptp(pha) <= 2 * np.pi, ("Your phase is probably in degrees "
                                          "and should be converted in radians "
                                          "using either np.degrees or "
                                          "np.deg2rad.")
        # Force the phase to be in [-pi, pi] :
        pha = (pha + np.pi) % (2. * np.pi) - np.pi
        return pha, amp

    def _infer_pvalues(self, effect, perm, p=.05, mcp='maxstat'):
        """Global function for statistical inferences.

        In order to work this method requires :

            * effect = array like of shape (n_dims...)
            * perm = array like of shape (n_perm, n_dims...)
        """
        assert all([isinstance(k, np.ndarray) for k in (effect, perm)])
        n_perm = perm.shape[0]
        # compute the minimum number of required permutations
        n_perm_req = int(10. / p)
        if n_perm < n_perm_req:
            logger.warning(f"For inferences at p<{p}, it is recommended to per"
                           f"form at least n_perm={n_perm_req} permutations")

        # ---------------------------------------------------------------------
        logger.info(f"    infer p-values at (p={p}, mcp={mcp})")
        # computes the pvalues
        if mcp is 'maxstat':
            max_p = perm.reshape(n_perm, -1).max(1)[np.newaxis, ...]
            nb_over = (effect[..., np.newaxis] <= max_p).sum(-1)
            pvalues = nb_over / n_perm
            # non-signi. p-values are set to 1. and min(pvalues) = 1 / n_perm
            pvalues[pvalues >= p] = 1.
            pvalues = np.maximum(1. / n_perm, pvalues)
        elif mcp in ['fdr', 'bonferroni']:
            from mne.stats import fdr_correction, bonferroni_correction
            fcn = fdr_correction if mcp is 'fdr' else bonferroni_correction
            # compute the p-values
            pvalues = (effect[np.newaxis, ...] <= perm).sum(0) / n_perm
            pvalues = np.maximum(1. / n_perm, pvalues)
            # apply correction
            is_signi, pvalues = fcn(pvalues, alpha=p)
            pvalues[~is_signi] = 1.

        return pvalues

    @property
    def f_pha(self):
        """Vector of phases of shape (n_pha, 2)."""
        return self._f_pha

    @property
    def f_amp(self):
        """Vector of amplitudes of shape (n_amp, 2)."""
        return self._f_amp

    @property
    def xvec(self):
        """Vector of phases of shape (n_pha,) use for plotting."""
        return self._xvec

    @property
    def yvec(self):
        """Vector of amplitudes of shape (n_amp,) use for plotting."""
        return self._yvec

    # ----------- DCOMPLEX -----------
    @property
    def dcomplex(self):
        """Get the dcomplex value."""
        return self._dcomplex

    @dcomplex.setter
    def dcomplex(self, value):
        """Set dcomplex value."""
        self._speccheck(dcomplex=value)

    # ----------- CYCLE -----------
    @property
    def cycle(self):
        """Get the cycle value."""
        return self._cycle

    @cycle.setter
    def cycle(self, value):
        """Set cycle value."""
        self._speccheck(cycle=value)

    # ----------- WIDTH -----------
    @property
    def width(self):
        """Get the width value."""
        return self._width

    @width.setter
    def width(self, value):
        """Set width value."""
        self._width = value


class Pac(_PacObj, _PacPlt):
    """Compute Phase-Amplitude Coupling (PAC).

    Computing PAC is assessed in three steps : compute the real PAC, compute
    surrogates and finally, because PAC is very sensible to the noise, correct
    the real PAC by the surrogates. This implementation is modular i.e. it lets
    you choose among a large range of possible combinations.

    Parameters
    ----------
    idpac : tuple/list | (1, 1, 3)
        Choose the combination of methods to use in order to extract PAC.
        This tuple must be composed of three integers where each one them
        refer

        * First digit : refer to the pac method

            - 1 : Mean Vector Length (MVL) :cite:`canolty2006high`
              (see :func:`tensorpac.methods.mean_vector_length`)
            - 2 : Modulation Index (MI) :cite:`tort2010measuring`
              (see :func:`tensorpac.methods.modulation_index`)
            - 3 : Heights Ratio (HR) :cite:`lakatos2005oscillatory`
              (see :func:`tensorpac.methods.heights_ratio`)
            - 4 : ndPAC :cite:`ozkurt2012statistically`
              (see :func:`tensorpac.methods.norm_direct_pac`)
            - 5 : Phase-Locking Value (PLV)
              :cite:`penny2008testing,lachaux1999measuring`
              (see :func:`tensorpac.methods.phase_locking_value`)
            - 6 : Gaussian Copula PAC (GCPAC) :cite:`ince2017statistical`
              (see :func:`tensorpac.methods.gauss_cop_pac`)

        * Second digit : refer to the method for computing surrogates

            - 0 : No surrogates
            - 1 : Swap phase / amplitude across trials
              :cite:`tort2010measuring`
              (see :func:`tensorpac.methods.swap_pha_amp`)
            - 2 : Swap amplitude time blocks
              :cite:`bahramisharif2013propagating`
              (see :func:`tensorpac.methods.swap_blocks`)
            - 3 : Time lag :cite:`canolty2006high`
              (see :func:`tensorpac.methods.time_lag`)

        * Third digit : refer to the normalization method for correction

            - 0 : No normalization
            - 1 : Substract the mean of surrogates
            - 2 : Divide by the mean of surrogates
            - 3 : Substract then divide by the mean of surrogates
            - 4 : Z-score

    f_pha, f_amp : list/tuple/array | def: [2, 4] and [60, 200]
        Frequency vector for the phase and amplitude. Here you can use
        several forms to define those vectors :

            * Basic list/tuple (ex: [2, 4] or [8, 12]...)
            * List of frequency bands (ex: [[2, 4], [5, 7]]...)
            * Dynamic definition : (start, stop, width, step)
            * Range definition (ex : np.arange(3) => [[0, 1], [1, 2]])
            * Using a string. `f_pha` and `f_amp` can be 'lres', 'mres', 'hres'
              respectively for low, middle and high resolution vectors. In that
              case, it uses the definition proposed by Bahramisharif et al.
              2013 :cite:`bahramisharif2013propagating` i.e
              f_pha = [f - f / 4, f + f / 4] and f_amp = [f - f / 8, f + f / 8]

    dcomplex : {'wavelet', 'hilbert'}
        Method for the complex definition. Use either 'hilbert' or
        'wavelet'.
    cycle : tuple | (3, 6)
        Control the number of cycles for filtering (only if dcomplex is
        'hilbert'). Should be a tuple of integers where the first one
        refers to the number of cycles for the phase and the second for the
        amplitude :cite:`bahramisharif2013propagating`.
    width : int | 7
        Width of the Morlet's wavelet.
    n_bins : int | 18
        Number of bins for the KLD and HR PAC method
        :cite:`tort2010measuring,lakatos2005oscillatory`
    """

    def __init__(self, idpac=(1, 2, 3), f_pha=[2, 4], f_amp=[60, 200],
                 dcomplex='hilbert', cycle=(3, 6), width=7, n_bins=18,
                 verbose=None):
        """Check and initialize."""
        set_log_level(verbose)
        self._idcheck(idpac)
        _PacObj.__init__(self, f_pha=f_pha, f_amp=f_amp, dcomplex=dcomplex,
                         cycle=cycle, width=width)
        _PacPlt.__init__(self)
        self.n_bins = int(n_bins)
        logger.info("Phase Amplitude Coupling object defined")

    def fit(self, pha, amp, n_perm=200, p=.05, mcp='maxstat', n_jobs=-1,
            random_state=None, verbose=None):
        """Compute PAC on filtered data.

        Parameters
        ----------
        pha : array_like
            Array of phases of shape (n_pha, n_epochs, n_times).
            Angles should be in rad.
        amp : array_like
            Array of amplitudes of shape (n_amp, n_epochs, n_times).
        n_perm : int | 200
            Number of surrogates to compute.
        p : float | 0.05
            Statistical threshold
        mcp : {'fdr', 'bonferroni'}
            Correct the p-values for multiple comparisons. Use either :

                * 'maxstat' : maximum statistics
                * 'fdr' : FDR correction (need MNE-Python)
                * 'bonferroni' : Bonferroni correction (need MNE-Python)
        n_jobs : int | -1
            Number of jobs to compute PAC in parallel. For very large data,
            set this parameter to 1 in order to prevent large memory usage.
        random_state : int | None
            Fix the random state of the machine for reproducible results.

        Returns
        -------
        pac : array_like
            Phase-Amplitude Coupling measure of shape (n_amp, n_pha, n_epochs)

        Attributes
        ----------
        pac : array_like
            Unormalized Phase-Amplitude Coupling measure of shape (n_amp,
            n_pha, n_epochs)
        pvalues : array_like
            Array of p-values of shape (n_amp, n_pha)
        surrogates : array_like
            Array of surrogates of shape (n_perm, n_amp, n_pha, n_epochs)
        """
        set_log_level(verbose)
        # ---------------------------------------------------------------------
        # input checking
        pha, amp = self._phampcheck(pha, amp)
        self._pvalues, self._surrogates = None, None
        # for the plv, extract the phase of the amplitude
        if self._idpac[0] == 5:
            amp = np.angle(hilbertm(amp))

        # ---------------------------------------------------------------------
        # check if permutations should be computed
        if self._idpac[1] == 0:
            n_perm = None
        if not isinstance(n_perm, int) or not (n_perm > 0):
            self._idpac = (self._idpac[0], 0, 0)
            compute_surro = False
        else:
            compute_surro = True

        # ---------------------------------------------------------------------
        # copnorm if gaussian copula is used
        if self._idpac[0] == 6:
            logger.debug(f"    copnorm the phase and the amplitude")
            pha = copnorm(np.stack([np.sin(pha), np.cos(pha)], axis=-2))
            amp = copnorm(amp[..., np.newaxis, :])

        # ---------------------------------------------------------------------
        # true pac estimation
        logger.info(f'    true PAC estimation using {self.method}')
        fcn = get_pac_fcn(self.idpac[0], self.n_bins, p)
        pac = fcn(pha, amp)
        self._pac = pac.copy()

        # ---------------------------------------------------------------------
        # compute surrogates (if needed)
        if compute_surro:
            if random_state is None:
                random_state = int(np.random.randint(0, 10000, size=1))
            logger.info(f"    compute surrogates ({self.str_surro}, {n_perm} "
                        f"permutations, random_state={random_state})")
            surro = compute_surrogates(pha, amp, self.idpac[1], fcn, n_perm,
                                       n_jobs, random_state)
            self._surrogates = surro

            # infer pvalues
            self.infer_pvalues(p, mcp=mcp)

        # ---------------------------------------------------------------------
        # normalize (if needed)
        if self._idpac[2] != 0:
            # Get the mean / deviation of surrogates
            logger.info("    normalize true PAC estimation by surrogates "
                        f"({self.str_norm})")
            normalize(self.idpac[2], pac, surro)

        return pac

    def filterfit(self, sf, x_pha, x_amp=None, n_perm=200, p=.05,
                  mcp='maxstat', edges=None, n_jobs=-1, random_state=None,
                  verbose=None):
        """Filt the data then compute PAC on it.

        Parameters
        ----------
        sf : float
            The sampling frequency.
        x_pha, x_amp : array_like
            Array of data for computing PAC. x_pha is the data used for
            extracting phases and x_amp, amplitudes. Both arrays must have
            the same shapes. If you want to compute PAC locally i.e. on the
            same electrode, x=x_pha=x_amp. For distant coupling, x_pha and
            x_amp could be different but still must to have the same shape.
        n_perm : int | 200
            Number of surrogates to compute.
        p : float | 0.05
            Statistical threshold
        mcp : {'fdr', 'bonferroni'}
            Correct the p-values for multiple comparisons. Use either :

                * 'maxstat' : maximum statistics
                * 'fdr' : FDR correction (need MNE-Python)
                * 'bonferroni' : Bonferroni correction (need MNE-Python)
        edges : int | None
            Number of samples to discard to avoid edge effects due to filtering
        n_jobs : int | -1
            Number of jobs to compute PAC in parallel. For very large data,
            set this parameter to 1 in order to prevent large memory usage.
        random_state : int | None
            Fix the random state of the machine for reproducible results.

        Returns
        -------
        pac : array_like
            Phase-Amplitude Coupling measure of shape (namp, npha, ...).

        Attributes
        ----------
        pac : array_like
            Unormalized Phase-Amplitude Coupling measure of shape (n_amp,
            n_pha, n_epochs)
        pvalues : array_like
            Array of p-values of shape (n_amp, n_pha)
        surrogates : array_like
            Array of surrogates of shape (n_perm, n_amp, n_pha, n_epochs)
        """
        # Check if amp is None :
        if x_amp is None:
            x_amp = x_pha
        # Shape checking :
        assert x_pha.shape == x_amp.shape, ("Inputs `x_pha` and `x_amp` must "
                                            "have the same shape.")
        # Extract phase (npha, ...) and amplitude (namp, ...) :
        logger.info(f"    extract phases (n_pha={len(self.xvec)}) and "
                    f"amplitudes (n_amps={len(self.yvec)})")
        kw = dict(keepfilt=False, edges=edges, n_jobs=1)
        pha = self.filter(sf, x_pha, 'phase', **kw)
        amp = self.filter(sf, x_amp, 'amplitude', **kw)

        # Special cases :
        if self._idpac[0] == 5:
            amp = np.angle(hilbertm(amp))

        # Compute pac :
        return self.fit(pha, amp, p=p, mcp=mcp, n_perm=n_perm, n_jobs=n_jobs,
                        random_state=random_state, verbose=verbose)

    def infer_pvalues(self, p=0.05, mcp='maxstat'):
        """Infer p-values based on surrogate distribution.

        Parameters
        ----------
        p : float | 0.05
            Significiency threshold

        Returns
        -------
        pvalues : array_like
            Array of p-values of shape (n_amp, n_pha)
        mcp : {'fdr', 'bonferroni'}
            Correct the p-values for multiple comparisons. Use either :

                * 'maxstat' : maximum statistics
                * 'fdr' : FDR correction (need MNE-Python)
                * 'bonferroni' : Bonferroni correction (need MNE-Python)
        """
        # ---------------------------------------------------------------------
        # check that pac and surrogates has already been computed
        assert hasattr(self, 'pac'), ("You should compute PAC first. Use the "
                                      "`fit` method")
        assert hasattr(self, 'surrogates'), "No surrogates computed"

        # mean pac and surrogates across trials
        m_pac, m_surro = self.pac.mean(2), self.surrogates.mean(3)
        self._pvalues = self._infer_pvalues(m_pac, m_surro, p=p, mcp=mcp)

        return self._pvalues

    def _idcheck(self, idpac):
        """Check the idpac parameter."""
        idpac = np.atleast_1d(idpac)
        if not all([isinstance(k, int) for k in idpac]) and (len(idpac) != 3):
            raise ValueError("idpac must be a tuple/list of 3 integers.")
        # Ozkurt PAC case (doesn't need surrogates and normalization)
        if idpac[0] == 4:
            idpac = np.array([4, 0, 0])
        if (idpac[0] == 1) and (idpac[1] == 0) and (idpac[2] == 0):
            logger.warning(
                "MVL is amplitude dependent which means that if the amplitude "
                "increases, MVL also increases. You should select a "
                "normalization method for correcting this limitation "
                "(e.g idpac=(1, 2, 4))")
        if (idpac[2] != 0) and (idpac[1] == 0):
            logger.warning("If you want to normalize the estimated PAC, you "
                           "should select a surrogate method (second digit of "
                           "`idpac`). Normalization ignored.")
            idpac[2] = 0
        self._idpac = idpac
        # string representation
        self.method, self.str_surro, self.str_norm = pacstr(idpac)

    @property
    def idpac(self):
        """Get the idpac value."""
        return self._idpac

    @idpac.setter
    def idpac(self, value):
        """Set idpac value."""
        self._idcheck(value)

    @property
    def pac(self):
        """Array of un-normalized PAC of shape (n_amp, n_pha, n_epochs)."""
        return self._pac

    @property
    def surrogates(self):
        """Array of surrogates of shape (n_perm, n_amp, n_pha, n_epochs)."""
        return self._surrogates

    @property
    def pvalues(self):
        """Array of p-values of shape (n_amp, n_pha)."""
        return self._pvalues


class EventRelatedPac(_PacObj, _PacVisual):
    """Compute the Event Related Phase-Amplitude Coupling (ERPAC).

    The traditional PAC approach is computed across time, hence this means that
    you can't observe PAC changes across time. In contrast, the ERPAC is
    computed across epochs (or trials) which preserves the time dimension.

    Parameters
    ----------
    f_pha, f_amp : list/tuple/array | def: [2, 4] and [60, 200]
        Frequency vector for the phase and amplitude. Here you can use
        several forms to define those vectors :

            * Basic list/tuple (ex: [2, 4] or [8, 12]...)
            * List of frequency bands (ex: [[2, 4], [5, 7]]...)
            * Dynamic definition : (start, stop, width, step)
            * Range definition (ex : np.arange(3) => [[0, 1], [1, 2]])

    dcomplex : {'wavelet', 'hilbert'}
        Method for the complex definition. Use either 'hilbert' or
        'wavelet'.
    cycle : tuple | (3, 6)
        Control the number of cycles for filtering (only if dcomplex is
        'hilbert'). Should be a tuple of integers where the first one
        refers to the number of cycles for the phase and the second for the
        amplitude.
    width : int | 7
        Width of the Morlet's wavelet.
    """

    def __init__(self, f_pha=[2, 4], f_amp=[60, 200], dcomplex='hilbert',
                 cycle=(3, 6), width=7, verbose=None):
        """Check and initialize."""
        set_log_level(verbose)
        _PacObj.__init__(self, f_pha=f_pha, f_amp=f_amp, dcomplex=dcomplex,
                         cycle=cycle, width=width)
        _PacPlt.__init__(self)
        logger.info("Event Related PAC object defined")

    def fit(self, pha, amp, method='circular', smooth=None, n_jobs=-1,
            n_perm=None, p=.05, mcp='fdr', verbose=None):
        """Compute the Event-Related Phase-Amplitude Coupling (ERPAC).

        The ERPAC :cite:`voytek2013method` is used to measure PAC across trials
        and is interesting for real-time estimation.

        Parameters
        ----------
        pha, amp : array_like
            Respectively the phase of slower oscillations of shape
            (n_pha, n_epochs, n_times) and the amplitude of faster
            oscillations of shape (n_pha, n_epochs, n_times).
        method : {'circular', 'gc'}
            Name of the method for computing erpac. Use 'circular' for
            reproducing :cite:`voytek2013method` or 'gc' for a Gaussian-Copula
            based erpac :cite:`ince2017statistical`.
        smooth : int | None
            Half number of time-points to use to produce a smoothing. Only
            active with the Gaussian-Copula ('gc') method.
        n_perm : int | None
            Number of permutations to compute for assessing p-values for the
            gaussian-copula ('gc') method. Statistics are performed by randomly
            swapping phase trials
        p : float | 0.05
            Statistical threshold for the gaussian-copula ('gc') method
        mcp : {'fdr', 'bonferroni'}
            Correct the p-values for multiple comparisons. This is needed when
            using the circular ERPAC (:cite:`voytek2013method`). Note that the
            correction is performed using MNE-Python.

        Returns
        -------
        erpac : array_like
            The ERPAC estimation of shape (n_amp, n_pha, n_times)
        """
        set_log_level(verbose)
        pha, amp = self._phampcheck(pha, amp)
        self.method = method
        self._pvalues = None
        # move the trial axis to the end (n_freqs, n_times, n_epochs)
        pha, amp = np.moveaxis(pha, 1, -1), np.moveaxis(amp, 1, -1)
        # method switch
        if method == 'circular':
            self.method = "ERPAC (Voytek et al. 2013)"
            logger.info(f"    Compute {self.method}")
            self._erpac, self._pvalues = erpac(pha, amp)
            self.infer_pvalues(p=p, mcp=mcp)
        elif method == 'gc':
            self.method = "Gaussian-Copula ERPAC"
            logger.info(f"    Compute {self.method}")
            # copnorm phases and amplitudes then compute erpac
            sco = copnorm(np.stack([np.sin(pha), np.cos(pha)], axis=-2))
            amp = copnorm(amp)[..., np.newaxis, :]
            self._erpac = ergcpac(sco, amp, smooth=smooth, n_jobs=n_jobs)
            # compute permutations (if needed)
            if isinstance(n_perm, int) and (n_perm > 0):
                logger.info(f"    Compute {n_perm} permutations")
                self._surrogates = _ergcpac_perm(sco, amp, smooth=smooth,
                                                 n_jobs=n_jobs, n_perm=n_perm)
                self.infer_pvalues(p=p, mcp=mcp)
        return self.erpac

    def filterfit(self, sf, x_pha, x_amp=None, method='circular', smooth=None,
                  n_perm=None, p=.05, mcp='fdr', edges=None, n_jobs=-1,
                  verbose=None):
        """Extract phases, amplitudes and compute ERPAC.

        Parameters
        ----------
        sf : float
            The sampling frequency.
        x_pha, x_amp : array_like
            Array of data for computing ERPAC. x_pha is the data used for
            extracting phases and x_amp, amplitudes. Both arrays must have
            the same shapes (i.e n_epochs, n_times). If you want to compute
            local ERPAC i.e. on the same electrode, x=x_pha=x_amp. For distant
            coupling, x_pha and x_amp could be different but still must to have
            the same shape.
        method : {'circular', 'gc'}
            Name of the method for computing erpac. Use 'circular' for
            reproducing :cite:`voytek2013method` or 'gc' for a Gaussian-Copula
            based erpac.
        smooth : int | None
            Half number of time-points to use to produce a smoothing. Only
            active with the Gaussian-Copula ('gc') method
            :cite:`ince2017statistical`.
        n_perm : int | None
            Number of permutations to compute for assessing p-values for the
            gaussian-copula ('gc') method. Statistics are performed by randomly
            swapping phase trials
        p : float | 0.05
            Statistical threshold for the gaussian-copula ('gc') method
        mcp : {'fdr', 'bonferroni'}
            Correct the p-values for multiple comparisons. This is needed when
            using the circular ERPAC (:cite:`voytek2013method`). Note that the
            correction is performed using MNE-Python.
        edges : int | None
            Number of samples to discard to avoid edge effects due to filtering

        Returns
        -------
        erpac : array_like
            The ERPAC estimation of shape (n_amp, n_pha, n_times)
        """
        x_amp = x_pha if not isinstance(x_amp, np.ndarray) else x_amp
        # extract phases and amplitudes
        logger.info(f"    Extract phases (n_pha={len(self.xvec)}) and "
                    f"amplitudes (n_amps={len(self.yvec)})")
        kw = dict(keepfilt=False, edges=edges, n_jobs=1)
        pha = self.filter(sf, x_pha, ftype='phase', **kw)
        amp = self.filter(sf, x_amp, ftype='amplitude', **kw)
        # compute erpac
        return self.fit(pha, amp, method=method, smooth=smooth, n_jobs=n_jobs,
                        n_perm=n_perm, p=p, mcp=mcp, verbose=verbose)

    def infer_pvalues(self, p=0.05, mcp='fdr'):
        """Infer p-values based on surrogate distribution.

        Parameters
        ----------
        p : float | 0.05
            Statistical threshold
        mcp : {'fdr', 'bonferroni'}
            Correct the p-values for multiple comparisons. This is needed when
            using the circular ERPAC (:cite:`voytek2013method`). Note that the
            correction is performed using MNE-Python.

        Returns
        -------
        pvalues : array_like
            Array of p-values of shape (n_amp, n_pha, n_times)
        """
        # ---------------------------------------------------------------------
        # check that pac and surrogates has already been computed
        assert hasattr(self, 'erpac'), ("You should compute ERPAC first. Use "
                                        "the `fit` method")
        assert mcp in ['fdr', 'bonferroni']

        # correct the p-values for multiple comparisons (Voytek's only)
        if "Voytek" in self.method:
            logger.info(f"    Correct p-values for multiple-comparisons using "
                        f"{mcp} correction of MNE-Python")
            from mne.stats import fdr_correction, bonferroni_correction
            fcn = fdr_correction if mcp is 'fdr' else bonferroni_correction
            _, self._pvalues = fcn(self._pvalues, alpha=p)
        else:
            assert hasattr(self, 'surrogates'), "No surrogates computed"
            # compute the p-values using maxstat (gcPAC)
            self._pvalues = self._infer_pvalues(self.erpac, self.surrogates,
                                                p=p)

        return self._pvalues

    @property
    def erpac(self):
        """Array of event-related PAC of shape ()."""
        return self._erpac

    @property
    def surrogates(self):
        """Array of surrogates of shape (n_perm, n_amp, n_pha, n_times)."""
        return self._surrogates

    @property
    def pvalues(self):
        """Array of p-values of shape (n_amp, n_pha, n_times)."""
        return self._pvalues


class PreferredPhase(_PacObj, _PolarPlt):
    """Compute the Preferred Phase (PP).

    The preferred phase is defined as the phase at which the amplitude is
    maximum.

    Parameters
    ----------
    f_pha, f_amp : list/tuple/array | def: [2, 4] and [60, 200]
        Frequency vector for the phase and amplitude. Here you can use
        several forms to define those vectors :

            * Basic list/tuple (ex: [2, 4] or [8, 12]...)
            * List of frequency bands (ex: [[2, 4], [5, 7]]...)
            * Dynamic definition : (start, stop, width, step)
            * Range definition (ex : np.arange(3) => [[0, 1], [1, 2]])

    dcomplex : {'wavelet', 'hilbert'}
        Method for the complex definition. Use either 'hilbert' or
        'wavelet'.
    cycle : tuple | (3, 6)
        Control the number of cycles for filtering (only if dcomplex is
        'hilbert'). Should be a tuple of integers where the first one
        refers to the number of cycles for the phase and the second for the
        amplitude.
    width : int | 7
        Width of the Morlet's wavelet.
    """

    def __init__(self, f_pha=[2, 4], f_amp=[60, 200], dcomplex='hilbert',
                 cycle=(3, 6), width=7, verbose=None):
        """Check and initialize."""
        set_log_level(verbose)
        _PacObj.__init__(self, f_pha=f_pha, f_amp=f_amp, dcomplex=dcomplex,
                         cycle=cycle, width=width)
        _PacPlt.__init__(self)
        logger.info("Preferred phase object defined")
        self.method = 'Preferred-Phase (PP)'

    def fit(self, pha, amp, n_bins=72):
        """Compute the preferred-phase.

        Parameters
        ----------
        pha, amp : array_like
            Respectively the phase of slower oscillations of shape
            (n_pha, n_epochs, n_times) and the amplitude of faster
            oscillations of shape (n_pha, n_epochs, n_times).
        n_bins : int | 72
            Number of bins for bining the amplitude according to phase
            slices.

        Returns
        -------
        binned_amp : array_like
            The binned amplitude according to the phase of shape
            (n_bins, n_amp, n_pha, n_epochs)
        pp : array_like
            The prefered phase where the amplitude is maximum of shape
            (namp, npha, n_epochs)
        polarvec : array_like
            The phase vector for the polar plot of shape (n_bins,)
        """
        # Check phase and amplitude shapes :
        pha, amp = self._phampcheck(pha, amp)
        return preferred_phase(pha, amp, n_bins=n_bins)

    def filterfit(self, sf, x_pha, x_amp=None, edges=None, n_bins=12,
                  verbose=None):
        """Extract phases, amplitudes and compute the preferred phase (PP).

        Parameters
        ----------
        sf : float
            The sampling frequency.
        x_pha, x_amp : array_like
            Array of data for computing PP. x_pha is the data used for
            extracting phases and x_amp, amplitudes. Both arrays must have
            the same shapes (i.e n_epochs, n_times). If you want to compute
            local PP i.e. on the same electrode, x=x_pha=x_amp. For distant
            coupling, x_pha and x_amp could be different but still must to have
            the same shape.
        n_bins : int | 72
            Number of bins for bining the amplitude according to phase
            slices.
        edges : int | None
            Number of samples to discard to avoid edge effects due to filtering

        Returns
        -------
        binned_amp : array_like
            The binned amplitude according to the phase of shape
            (n_bins, n_amp, n_pha, n_epochs)
        pp : array_like
            The prefered phase where the amplitude is maximum of shape
            (namp, npha, n_epochs)
        polarvec : array_like
            The phase vector for the polar plot of shape (n_bins,)
        """
        x_amp = x_pha if not isinstance(x_amp, np.ndarray) else x_amp
        # extract phases and amplitudes
        logger.info(f"    Extract phases (n_pha={len(self.xvec)}) and "
                    f"amplitudes (n_amps={len(self.yvec)})")
        kw = dict(keepfilt=False, edges=edges, n_jobs=1)
        pha = self.filter(sf, x_pha, ftype='phase', **kw)
        amp = self.filter(sf, x_amp, ftype='amplitude', **kw)
        # compute pp
        return self.fit(pha, amp, n_bins=n_bins)
